package de.dfki.km.academicresearch.eswc2016challenge.pathfinder.jade;

import de.dfki.km.academicresearch.eswc2016challenge.data.Path;
import de.dfki.km.academicresearch.eswc2016challenge.pathfinder.basis.PathFinder;
import de.dfki.km.jade.graph.core.CorePath;
import de.dfki.km.jade.graph.multisteinertree.MultiSteinerTree;
import de.dfki.km.jade.graph.search.ShrinkingWeightCalculator;
import de.dfki.km.jade.graph.search.SubGraph;
import de.dfki.km.jade.graph.search.WalkedPathsInterruptConfig;
import de.dfki.km.jade.graph.search.WeightCalculator;
import de.dfki.km.jade.util.FastReadArrayList;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.SortedSet;
import org.apache.jena.graph.Triple;
import org.apache.jena.rdf.model.Resource;
import org.apache.jena.rdf.model.ResourceFactory;
import org.nustaq.serialization.FSTObjectInput;
import org.nustaq.serialization.FSTObjectOutput;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 *
 * @author Markus Schr&ouml;der
 */
public class JadePathFinder extends PathFinder {

    private GraphIndexJade jade;

    private static final Logger logger = LoggerFactory.getLogger(JadePathFinder.class);

    public JadePathFinder() {
        jade = new GraphIndexJade();
    }

    @Override
    public List<Path> find(String srcURI, String dstURI, int k, String startOrEndProp) throws Exception {

        MultiSteinerTree<SPOVertex, SPOEdge> steinerTree = toSteinerTree(k);

        List<Set<SPOVertex>> startNodeSets = toStartNodeSets(srcURI, dstURI);

        SortedSet<SubGraph<SPOVertex, SPOEdge>> results = search(startNodeSets, steinerTree);

        return toPathes(results);
    }

    private MultiSteinerTree<SPOVertex, SPOEdge> toSteinerTree(int k) {
        int resultCount4Interrupt = k;
        int walkedPathsCount4Interrupt = 10000;

        WeightCalculator<SPOVertex, SPOEdge, SubGraph<SPOVertex, SPOEdge>> wc = new ShrinkingWeightCalculator<>();
        ((ShrinkingWeightCalculator) wc).setFilterDoubleNodeEntryGraphs(true);

        MyWeightCalculator mywc = new MyWeightCalculator();

        WalkedPathsInterruptConfig ic = new WalkedPathsInterruptConfig(resultCount4Interrupt, walkedPathsCount4Interrupt);

        MultiSteinerTree<SPOVertex, SPOEdge> steinerTree = new MultiSteinerTree<>(jade.getGraph(), mywc, ic);//only a constructor

        //Filter out subgraphs with same node set
        steinerTree.setResultEqualityWithoutRoots(true);//only an attribute is set

        return steinerTree;
    }

    private List<Set<SPOVertex>> toStartNodeSets(String srcURI, String dstURI) throws Exception {
        List<Set<SPOVertex>> startNodeSets = new LinkedList<>();

        for (String id : Arrays.asList(srcURI, dstURI)) {
            Set<SPOVertex> startNodes = new HashSet<>();
            SPOVertex vertex = jade.getId2Vert().get(id);
            startNodes.add(vertex);
            startNodeSets.add(startNodes);
        }

        if (startNodeSets.size() <= 1) {
            throw new Exception(startNodeSets.size() + " start node size problem");
        }

        return startNodeSets;
    }

    private SortedSet<SubGraph<SPOVertex, SPOEdge>> search(List<Set<SPOVertex>> startNodeSets, MultiSteinerTree<SPOVertex, SPOEdge> steinerTree) throws Exception {
        //search: here happens the processing
        SortedSet<SubGraph<SPOVertex, SPOEdge>> steinerResult = null;
        try {
            steinerResult = steinerTree.search(startNodeSets);
        } catch (Exception e) {
            throw new Exception("", e);
        }

        //empty result
        if (steinerResult.isEmpty()) {
            throw new Exception("");
        }

        return steinerResult;
    }

    private List<Path> toPathes(SortedSet<SubGraph<SPOVertex, SPOEdge>> results) {

        /*
        
        ["http://dbpedia.org/resource/Felipe_Massa", "http://dbpedia.org/property/after", "http://dbpedia.org/resource/Robert_Kubica", "http://dbpedia.org/property/firstWin", "http://dbpedia.org/resource/2008_Canadian_Grand_Prix", "http://dbpedia.org/property/thirdTeam", "http://dbpedia.org/resource/Red_Bull"]
        ["http://dbpedia.org/resource/Felipe_Massa", "http://dbpedia.org/property/after", "http://dbpedia.org/resource/Robert_Kubica", "http://dbpedia.org/ontology/firstWin", "http://dbpedia.org/resource/2008_Canadian_Grand_Prix", "http://dbpedia.org/property/thirdTeam", "http://dbpedia.org/resource/Red_Bull"]
        ["http://dbpedia.org/resource/Felipe_Massa", "http://dbpedia.org/property/after", "http://dbpedia.org/resource/Robert_Kubica", "http://dbpedia.org/ontology/lastWin", "http://dbpedia.org/resource/2008_Canadian_Grand_Prix", "http://dbpedia.org/ontology/thirdTeam", "http://dbpedia.org/resource/Red_Bull"]
        ["http://dbpedia.org/resource/Felipe_Massa", "http://dbpedia.org/property/after", "http://dbpedia.org/resource/Robert_Kubica", "http://dbpedia.org/property/lastWin", "http://dbpedia.org/resource/2008_Canadian_Grand_Prix", "http://dbpedia.org/ontology/thirdTeam", "http://dbpedia.org/resource/Red_Bull"]
        ["http://dbpedia.org/resource/Felipe_Massa", "http://dbpedia.org/property/after", "http://dbpedia.org/resource/Robert_Kubica", "http://dbpedia.org/property/firstWin", "http://dbpedia.org/resource/2008_Canadian_Grand_Prix", "http://dbpedia.org/ontology/thirdTeam", "http://dbpedia.org/resource/Red_Bull"]
        ["http://dbpedia.org/resource/Felipe_Massa", "http://dbpedia.org/property/after", "http://dbpedia.org/resource/Robert_Kubica", "http://dbpedia.org/ontology/lastWin", "http://dbpedia.org/resource/2008_Canadian_Grand_Prix", "http://dbpedia.org/property/thirdTeam", "http://dbpedia.org/resource/Red_Bull"]
        ["http://dbpedia.org/resource/Felipe_Massa", "http://dbpedia.org/property/after", "http://dbpedia.org/resource/Robert_Kubica", "http://dbpedia.org/property/lastWin", "http://dbpedia.org/resource/2008_Canadian_Grand_Prix", "http://dbpedia.org/property/thirdTeam", "http://dbpedia.org/resource/Red_Bull"]
        ["http://dbpedia.org/resource/Felipe_Massa", "http://dbpedia.org/property/after", "http://dbpedia.org/resource/Robert_Kubica", "http://dbpedia.org/ontology/firstWin", "http://dbpedia.org/resource/2008_Canadian_Grand_Prix", "http://dbpedia.org/ontology/thirdTeam", "http://dbpedia.org/resource/Red_Bull"]

         */
        List<Path> pathes = new LinkedList<>();

        for (SubGraph<SPOVertex, SPOEdge> subGraph : results) {

            logger.debug(subGraph.toString());

            List<Resource> resourceList = new LinkedList<>();

            for (CorePath<SPOVertex, SPOEdge> path : subGraph.getPaths()) {

                logger.debug("\t" + path.toString());

                List<SPOEdge> edges = pathGetEdges(path);
                for (SPOEdge edge : edges) {

                    logger.debug("\t\t" + edge.toString());

                    Resource s = ResourceFactory.createResource(edge.getSource().getId());
                    Resource p = ResourceFactory.createProperty(edge.getPredicate());
                    //Resource o = ResourceFactory.createResource(edge.getTarget().getId());

                    resourceList.add(s);
                    resourceList.add(p);
                }

                //last object
                if (!edges.isEmpty()) {
                    SPOEdge lastEdge = edges.get(edges.size() - 1);
                    resourceList.add(ResourceFactory.createResource(lastEdge.getTarget().getId()));
                }
            }

            //make a path with resource list
            pathes.add(new Path(resourceList));
        }

        return pathes;
    }

    private List<SPOEdge> pathGetEdges(CorePath<SPOVertex, SPOEdge> path) {
        FastReadArrayList<SPOEdge> edges = new FastReadArrayList<>(path.getLength() - 1);

        CorePath<SPOVertex, SPOEdge> currentPathCursor = path;

        // im ersten PathCursor gibt es keine Edge (startNode==endNode)
        for (int i = path.getLength() - 2; i >= 0; i--) {
            SPOEdge e = currentPathCursor.getEndEdge();

            //if(!currentPathCursor.isEndEdgeDirectionForward())
            //    e = new SPOEdge(e.getTarget(), e.getPredicate(), e.getSource());
            edges.setQuick(i, e);

            currentPathCursor = currentPathCursor.getParent();
        }

        return edges;
    }

    private class MyWeightCalculator extends WeightCalculator<SPOVertex, SPOEdge, SubGraph<SPOVertex, SPOEdge>> {

        private WeightCalculator<SPOVertex, SPOEdge, SubGraph<SPOVertex, SPOEdge>> defaultWC;

        public MyWeightCalculator() {
            defaultWC = new ShrinkingWeightCalculator<>();
            ((ShrinkingWeightCalculator) defaultWC).setFilterDoubleNodeEntryGraphs(true);
        }

        @Override
        public float calculateWeight(SPOVertex newEndVertex, SPOEdge newEndEdge, CorePath<SPOVertex, SPOEdge> parentPath) {

            //logger.debug(parentPath.toString());
            //logger.debug(newEndEdge.toString());
            //logger.debug(newEndVertex.getId());
            /*
            if(newEndEdge.getPredicate().equals("http://www.w3.org/1999/02/22-rdf-syntax-ns#type"))
                return 2.0f + parentPath.getWeight();
             */
            return 1f + parentPath.getWeight();
        }

        @Override
        public boolean filterResult(SubGraph<SPOVertex, SPOEdge> potentialResult, Set<SubGraph<SPOVertex, SPOEdge>> currentResults) {
            //never filter anything
            //return false;

            //use ShrinkingWeightCalculator for this job
            try {
                return defaultWC.filterResult(potentialResult, currentResults);
            } catch (java.lang.ArrayIndexOutOfBoundsException ex) {
                return true;
            }
        }

    }

//================================================
//================================================
//================================================
//================================================
    //INDEX Specifica
    private ArrayList<SPOVertex> vs = new ArrayList<>(367772);
    private ArrayList<SPOEdge> es = new ArrayList<>(2397994);
    private int i = 0;
    private int literalCount = 0;

    @Override
    public void send(Triple t) {
        //logger.debug(t.toString());

        if (t.getObject().isLiteral()) {
            literalCount++;
            return;
        }

        SPOVertexLookup sl = getVertex(t.getSubject().getURI(), jade.getId2Vert());
        SPOVertexLookup ol = getVertex(t.getObject().getURI(), jade.getId2Vert());

        if (!sl.found) {
            vs.add(sl.vertex);
        }

        if (!ol.found) {
            vs.add(ol.vertex);
        }

        SPOEdge e = new SPOEdge(sl.vertex, t.getPredicate().getURI(), ol.vertex);
        es.add(e);

        if (i > 0 && i % 100000 == 0) {
            logger.debug("processed: " + i);
        }
        i++;

    }

    @Override
    public void endStreaming() throws Exception {
        jade.load(vs, es);
        logger.debug("flush " + jade.toString() + ", " + literalCount + " literals skipped");
        File saveFile = getIndexFile();

        logger.debug("write to " + saveFile.getAbsolutePath());
        /*
        ObjectOutputStream out = null;
        try {
            out = new ObjectOutputStream(new FileOutputStream(saveFile));
            out.writeObject(vs);
            out.writeObject(es);
            out.close();
        } catch (IOException ex) {
            logger.error(null, ex);
        } finally {
            try {
                out.close();
            } catch (IOException ex) {
                logger.error(null, ex);
            }
        }
         */
        FSTObjectOutput out = null;
        try {
            out = new FSTObjectOutput(new FileOutputStream(saveFile));
            out.writeObject(vs);
            out.writeObject(es);
        } finally {
            out.close();
        }

        logger.debug("done");

        //27048 [main] DEBUG de.dfki.km.academicresearch.eswc2016challenge.Test.JadeSink2  - write to vs_es.bin
        //45227 [main] DEBUG de.dfki.km.academicresearch.eswc2016challenge.Test.JadeSink2  - done
        //183 MB file		
    }

    ;

	@Override
    public void loadIndex() throws Exception {

        ArrayList<SPOVertex> tmpvs;
        ArrayList<SPOEdge> tmpes;

        File loadFile = getIndexFile();

        /*
        ObjectInputStream out = new ObjectInputStream(new FileInputStream(loadFile));
        logger.debug("start deserialize");
        
        tmpvs  = (ArrayList<SPOVertex>) out.readObject();
        tmpes = (ArrayList<SPOEdge>) out.readObject();
        
        logger.debug("done");
        out.close();
         */
        FSTObjectInput out = null;
        try {
            out = new FSTObjectInput(new FileInputStream(loadFile));
            logger.debug("start deserialize");
            tmpvs = (ArrayList<SPOVertex>) out.readObject();
            tmpes = (ArrayList<SPOEdge>) out.readObject();
            logger.debug("done");
        } finally {
            out.close();
        }

        jade.load(tmpvs, tmpes);

        for (SPOVertex v : tmpvs) {
            jade.getId2Vert().put(v.getId(), v);
        }

    }

    private final int NEW = -1;

    private SPOVertexLookup getVertex(String id, Map<String, SPOVertex> id2vert) {
        if (id2vert.containsKey(id)) {
            return new SPOVertexLookup(id2vert.get(id), true);
        } else {
            SPOVertex v = new SPOVertex(id);
            v.setIndex(NEW);//to mark it as new
            id2vert.put(v.getId(), v);
            return new SPOVertexLookup(v, false);
        }
    }

    private class SPOVertexLookup {

        public SPOVertex vertex;
        public boolean found;

        public SPOVertexLookup(SPOVertex vertex, boolean found) {
            this.vertex = vertex;
            this.found = found;
        }
    }
}
