package de.dfki.km.academicresearch.eswc2016challenge.pathfinder.jgraphturiedge;

import de.dfki.km.academicresearch.eswc2016challenge.data.Path;
import de.dfki.km.academicresearch.eswc2016challenge.pathfinder.jgraphturiedge.base.JGraphTFinderBaseURI;
import de.dfki.km.academicresearch.eswc2016challenge.pathfinder.jgraphturiedge.base.URIEdge;
import de.dfki.km.academicresearch.eswc2016challenge.pathfinder.jgraphturiedge.shortestpathtree.BFSDistance;
import de.dfki.km.academicresearch.eswc2016challenge.pathfinder.jgraphturiedge.shortestpathtree.IShortestPathTree;
import de.dfki.km.academicresearch.eswc2016challenge.pathfinder.jgraphturiedge.shortestpathtree.ShortestPathTree;
import java.util.Collections;
import java.util.Comparator;
import java.util.LinkedList;
import java.util.List;
import java.util.Objects;
import java.util.Queue;
import java.util.Set;
import java.util.Stack;
import static java.util.stream.Collectors.toList;
import org.apache.jena.rdf.model.ResourceFactory;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 *
 * @author Markus Schr&ouml;der
 */
public class PartialPathsFinder extends JGraphTFinderBaseURI {
    
    private static final Logger logger = LoggerFactory.getLogger(PartialPathsFinder.class);
    
    @Override
    public List<Path> find(String srcURI, String dstURI, int k, String startOrEndProp) throws Exception {
        return findKShortestPath(srcURI, dstURI, k, startOrEndProp);
    }

    protected List<Path> findKShortestPath(String srcURI, String dstURI, int k, String startOrEndProp) throws Exception {
        
        //shortest path tree
        IShortestPathTree<String, URIEdge> sptCalculator = new BFSDistance<>();
        ShortestPathTree<String, URIEdge> spt = sptCalculator.computeShortestPathTree(g, dstURI);
        
        //data storage
        Queue<URIEdgePath> next = new LinkedList<>();
        List<URIEdgePath> found = new LinkedList<>();

        //init with srcURI
        next.add(new URIEdgePath(new URIEdge("", "", srcURI)));
        //later we remove it in toPath in URIEdgePath
        
        //BFS
        //for each path we go
        while (!next.isEmpty() && found.size() < k) {

            //the path we go currently
            URIEdgePath partial = next.poll();

            logger.debug("partial: {}", partial.toString());
            
            //we use shortest path tree to get directly a complete shortest path
            List<URIEdgePath> completes = completeMulti(partial, dstURI, spt);

            //is a real complete path and no duplicate
            for(URIEdgePath complete : completes) {
                if(!found.contains(complete)) {
                    logger.debug("complete: {}", complete.toPath());
                    found.add(complete);
                }
            }
            
            //next pathes
            String last = partial.getLast();
            logger.debug("last: {}", last);
            
            //we look at the branches
            Set<URIEdge> outs = g.outgoingEdgesOf(last);
            
            logger.debug("outs: {} edges", outs.size());
            
            for (URIEdge e : outs) {
                
                //no duplicate triples in our current path
                if(partial.path.contains(e))
                    continue;
                
                //concat the current with the new statement
                URIEdgePath newPath = new URIEdgePath(partial, e);
                next.add(newPath);
            }
            
            //we should take the sidetracks first that have short distance to end node
            Collections.sort((List<URIEdgePath>) next, new Comparator<URIEdgePath>() {
                @Override
                public int compare(URIEdgePath a, URIEdgePath b) {
                    String c = a.getLast();
                    String d = b.getLast();
                    
                    Double x = spt.getDistance(c);
                    Double y = spt.getDistance(d);
                    
                    if(x == null && y == null) {
                        return 0;
                    }
                    
                    if(x == null) {
                        return 1;
                    }
       
                    if(y == null) {
                        return -1;
                    }
                    
                    return Double.compare(x, y);
                }
            });
            
            logger.debug("");
        }

        return found.stream().map(p -> p.toPath()).collect(toList());
    }

    protected URIEdgePath complete(URIEdgePath e, ShortestPathTree<String, URIEdge> spt) {
        URIEdgePath complete = new URIEdgePath(e);
        
        String cur = complete.getLast();
        while(true) {
            URIEdge tmp = spt.getPrevious(cur);
            if(tmp == null)
                break;
            
            complete.path.add(tmp);
            
            cur = tmp.getTarget();
        }
        
        return complete;
    }
    
    protected List<URIEdgePath> completeMulti(URIEdgePath e, String dstURI, ShortestPathTree<String, URIEdge> spt) {
        //Queue<URIEdgePath> q = new LinkedList<>();
        Stack<URIEdgePath> s = new Stack<>();
        
        //q.add(new URIEdgePath(e));
        s.push(new URIEdgePath(e));
        
        List<URIEdgePath> found = new LinkedList<>();
        
        while(!s.isEmpty()) {
            
            URIEdgePath r = s.pop();

            logger.debug("r: {}", r.toString());
            
            String cur = r.getLast();

            //found
            if(cur.equals(dstURI)) {
                found.add(r);
                continue;
            }
            
            Set<URIEdge> tmps = spt.getPreviousMulti(cur);
            
            //no way to target
            if(tmps == null)
                continue; //partial path dies

            //all multi edges
            for(URIEdge tmp : tmps) {
            
                logger.debug("tmp(multi): {}", tmp);
                
                //new complete path per edge of multi edges
                URIEdgePath newPath = new URIEdgePath(r, tmp);
                s.push(newPath);
            }
        }
        
        return found;
    }
    
    protected class URIEdgePath implements Comparable<URIEdgePath> {

        public List<URIEdge> path;

        public URIEdgePath() {
            this.path = new LinkedList<>();
        }

        public URIEdgePath(URIEdge e) {
            this.path = new LinkedList<>();
            path.add(e);
        }

        public URIEdgePath(URIEdgePath path) {
            this.path = new LinkedList<>(path.path);
        }
        
        public URIEdgePath(URIEdgePath path, URIEdge newE) {
            this.path = new LinkedList<>(path.path);
            this.path.add(newE);
        }

        public Path toPath() {
            Path p = new Path();
            List<URIEdge> internal = path.subList(1, path.size());
            if (!internal.isEmpty()) {
                p.add(ResourceFactory.createResource(internal.get(0).getSource()));
            }
            for (URIEdge e : internal) {
                p.add(ResourceFactory.createProperty(e.getEdge()));
                p.add(ResourceFactory.createResource(e.getTarget()));
            }
            return p;
        }
        
        public String getLast() {
            if (path.isEmpty()) {
                return null;
            }
            return path.get(path.size() - 1).getTarget();
        }
        
        @Override
        public String toString() {
            StringBuilder sb = new StringBuilder();

            sb.append("(");
            sb.append(path.size());
            sb.append(") ");

            for (URIEdge e : path) {
                sb.append(e.toString());
            }

            return sb.toString();
        }

        @Override
        public int compareTo(URIEdgePath other) {
            return Integer.compare(this.path.size(), other.path.size());
        }

        @Override
        public int hashCode() {
            int hash = 3;
            hash = 19 * hash + Objects.hashCode(this.path);
            return hash;
        }

        @Override
        public boolean equals(Object obj) {
            if (this == obj) {
                return true;
            }
            if (obj == null) {
                return false;
            }
            if (getClass() != obj.getClass()) {
                return false;
            }
            final URIEdgePath other = (URIEdgePath) obj;
            if (!Objects.equals(this.path, other.path)) {
                return false;
            }
            return true;
        }

    }

}
