package de.dfki.km.academicresearch.eswc2016challenge.sparqlext.dataset;

import java.util.Iterator;
import java.util.List;

import org.apache.jena.graph.Graph;
import org.apache.jena.graph.Node;
import org.apache.jena.query.ReadWrite;
import org.apache.jena.query.text.TextDocProducer;
import org.apache.jena.query.text.TextHit;
import org.apache.jena.query.text.TextIndex;
import org.apache.jena.query.text.TextIndexException;
import org.apache.jena.sparql.core.DatasetGraph;
import org.apache.jena.sparql.core.DatasetGraphMonitor;
import org.apache.jena.sparql.core.DatasetGraphWithLock;
import org.apache.jena.sparql.core.GraphView;
import org.apache.jena.sparql.core.Transactional;
import org.apache.lucene.queryparser.classic.QueryParserBase;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import de.dfki.km.academicresearch.eswc2016challenge.data.Path;

public class DatasetGraphGraph extends DatasetGraphMonitor implements Transactional {

	private static Logger       log = LoggerFactory.getLogger(DatasetGraphGraph.class) ;
    private final GraphIndex    graphIndex ;
    private final Transactional dsgtxn ;
    private final Graph         dftGraph ;
    private final boolean       closeIndexOnClose;
    
    
    
    // If we are going to implement Transactional, then we are going to have to do as DatasetGraphWithLock and
    // TDB's DatasetGraphTransaction do and track transaction state in a ThreadLocal
    private final ThreadLocal<ReadWrite> readWriteMode = new ThreadLocal<ReadWrite>();
    
    
    public DatasetGraphGraph(DatasetGraph dsg, GraphIndex index, GraphSynchronizer producer)
    { 
        this(dsg, index, producer, false);
    }
    
    public DatasetGraphGraph(DatasetGraph dsg, GraphIndex index, GraphSynchronizer producer, boolean closeIndexOnClose)
    {
        super(dsg, producer) ;
        this.graphIndex = index ;
        if ( dsg instanceof Transactional )
            dsgtxn = (Transactional)dsg ;
        else
            dsgtxn = new DatasetGraphWithLock(dsg) ;
        dftGraph = GraphView.createDefaultGraph(this) ;
        this.closeIndexOnClose = closeIndexOnClose;
    }
    
    
    
 // ---- Intercept these and force the use of views.
    @Override
    public Graph getDefaultGraph() {
        return dftGraph ;
    }

    @Override
    public Graph getGraph(Node graphNode) {
        return GraphView.createNamedGraph(this, graphNode) ;
    }

    // ----

    public GraphIndex getGraphIndex() {
        return graphIndex ;
    }
    
    
    
    
    /** Search the text index on the default text field */
	public Iterator<Path> find(String srcURI, String dstURI, int k, String startOrEndProp) throws Exception {
		return graphIndex.find(srcURI, dstURI, k, startOrEndProp).iterator();
	}
    

    @Override
    public void begin(ReadWrite readWrite) {
        readWriteMode.set(readWrite);
        dsgtxn.begin(readWrite) ;
        super.getMonitor().start() ;
    }
    
    /**
     * Rollback all changes, discarding any exceptions that occur.
     */
    @Override
    public void abort() {
        // Roll back all both objects, discarding any exceptions that occur
        try { dsgtxn.abort(); } catch (Throwable t) { log.warn("Exception in abort: " + t.getMessage(), t); }
        try { graphIndex.rollback(); } catch (Throwable t) { log.warn("Exception in abort: " + t.getMessage(), t); }
        readWriteMode.set(null) ;
        super.getMonitor().finish() ;
    }
    
    /**
     * Perform a 2-phase commit by first calling prepareCommit() on the GraphIndex
     * followed by committing the Transaction object, and then calling commit()
     * on the GraphIndex().
     * <p> 
     * If either of the objects fail on either the preparation or actual commit,
     * it terminates and calls {@link #abort()} on both of them.
     * <p>
     * <b>NOTE:</b> it may happen that the GraphIndex fails to commit, after the
     * Transactional has already successfully committed.  A rollback instruction will
     * still be issued, but depending on the implementation, it may not have any effect.
     */
    @Override
    public void commit() {
        super.getMonitor().finish() ;
        // Phase 1
        if (readWriteMode.get() == ReadWrite.WRITE) {
            try {
            	graphIndex.prepareCommit();
            }
            catch (Throwable t) {
                log.error("Exception in prepareCommit: " + t.getMessage(), t) ;
                abort();
                throw new TextIndexException(t);
            }
        }
        
        // Phase 2
        try {
        	dsgtxn.commit();
            if (readWriteMode.get() == ReadWrite.WRITE) {
            	graphIndex.commit();
            }
        }
        catch (Throwable t) {
            log.error("Exception in commit: " + t.getMessage(), t) ;
            abort();
            throw new TextIndexException(t);
        }
        readWriteMode.set(null);
    }

    @Override
    public boolean isInTransaction() {
        return readWriteMode.get() != null;
    }

    @Override
    public void end() {
        // If we are still in a write transaction at this point, then commit was never called, so rollback the GraphIndex
        if (readWriteMode.get() == ReadWrite.WRITE) {
            try {
            	graphIndex.rollback();
            }
            catch (Throwable t) {
                log.warn("Exception in end: " + t.getMessage(), t) ;
            }
        }
        
        try {
        	dsgtxn.end() ;
        }
        catch (Throwable t) {
            log.warn("Exception in end: " + t.getMessage(), t) ;
        }
        
        readWriteMode.set(null) ;
        super.getMonitor().finish() ;
    }
    
    @Override
    public void close() {
        super.close();
        if (closeIndexOnClose) {
        	graphIndex.close();
        }
    }
    
    
}
