package de.dfki.km.jade.graph;





/**
 * Represents a part of a graph with one defined start and one defined end. This corresponds to a {@link HierarchicalGraph} with one leaf. Thus, it
 * has a start, an ende, a length and an edge direction
 * 
 * @author reuschling
 * 
 * @param <V> the vertex type
 * @param <E> the edge type
 */
public interface Path<V extends Vertex, E extends Edge<V>> extends HierarchicalGraph<V, E>
{
    /**
     * Gets the last (end/leaf) edge of this path
     * 
     * @return the last (end/leaf) edge of this path
     */
    public E getEndEdge();



    /**
     * Same as {@link #getLeafs()}[0]
     * 
     * @return the last (end/leaf) node of this path
     */
    public V getEndVertex();



    /**
     * Gets the length of this path. This corresponds to the count of edges
     * 
     * @return the length of this path. This corresponds to the count of edges
     */
    public int getLength();



    /**
     * Gets the first (start/root) edge of this path
     * 
     * @return the first (start/root) edge of this path
     */
    public E getStartEdge();



    /**
     * Same as {@link #getRoot()}
     * 
     * @return the first (start/root) node of this path
     */
    public V getStartVertex();



    /**
     * Gives the information if an edge, as part of this path, goes from left to right (forward), where the direction goes from source node to target
     * node - or vice versa. You can also get this information by looking to the source/target vertex and look to the next/last edge inside this path.
     * 
     * @param indexOfEdgeInPath the position of the edge inside this path. (one edge could be more than once inside the path, so we use an index-based
     *            method)
     * 
     * @return true if the edge is a forward-edge, thus it goes from left to right (source=>target). False otherwise.
     */
    public boolean isEdgeDirectionForward(int indexOfEdgeInPath);


}
