package de.dfki.km.jade.graph.core;



import gnu.trove.list.array.TIntArrayList;

import java.util.Collection;
import java.util.LinkedList;
import java.util.List;
import java.util.logging.Logger;

import de.dfki.km.jade.graph.BiAdjacencyList;
import de.dfki.km.jade.graph.Edge;
import de.dfki.km.jade.graph.Graph;
import de.dfki.km.jade.graph.Vertex;
import de.dfki.km.jade.util.FastReadArrayList;
import de.dfki.km.jade.util.StringUtils;



public class CoreBiAdjacencyList<V extends Vertex, E extends Edge<V>> implements BiAdjacencyList<V, E>
{

    protected FastReadArrayList<FastReadArrayList<E>> m_fraEdges4SourceIndex = new FastReadArrayList<>();

    protected FastReadArrayList<FastReadArrayList<E>> m_fraEdges4TargetIndex = new FastReadArrayList<>();



    public CoreBiAdjacencyList()
    {
    }







    public CoreBiAdjacencyList(Collection<E> edges)
    {
        setContent(edges, false);
    }



    public CoreBiAdjacencyList(Collection<E> edges, boolean bVerbose)
    {
        setContent(edges, bVerbose);
    }




    public CoreBiAdjacencyList(Graph<V, E> graph)
    {
        this(graph.getEdges());
    }





    /**
     * <b>CAUTION : </b>This is a very costly method, don't use it to create the whole BiAdjacencyList, only for slight modifications. The internal array structures will
     * be grown on every invocation with arraycopies. This design is in favour of read performance and memory consumtion against incremental write performance. To create
     * the whole AdjacencyList fast, use the constructors or {@link #setContent(Collection)}.
     */
    @Override
    public void addEdge(E newEdge)
    {

        Vertex sourceVertex = newEdge.getSource();
        Vertex targetVertex = newEdge.getTarget();


        int iMaxIndex = Math.max(sourceVertex.getIndex(), targetVertex.getIndex());
        m_fraEdges4SourceIndex.grow(iMaxIndex + 1);
        m_fraEdges4TargetIndex.grow(iMaxIndex + 1);



        // die targets für die source
        FastReadArrayList<E> edgesWithCurrentSourceVertexSource = m_fraEdges4SourceIndex.get(sourceVertex.getIndex());

        if(edgesWithCurrentSourceVertexSource == null)
        {
            edgesWithCurrentSourceVertexSource = new FastReadArrayList<>();

            m_fraEdges4SourceIndex.setQuick(sourceVertex.getIndex(), edgesWithCurrentSourceVertexSource);
        }

        // das 'add' hier ist etwas unglücklich, da hier immer nur um 1 erweitert wird...vielleicht doch mit ArrayLists und einer trim()-Methode?
        // Wird dann allerdings sehr schnell recht speicherintensiv...
        edgesWithCurrentSourceVertexSource.add(newEdge);




        FastReadArrayList<E> edgesWithCurrentTargetVertexTarget = m_fraEdges4TargetIndex.get(targetVertex.getIndex());

        if(edgesWithCurrentTargetVertexTarget == null)
        {
            edgesWithCurrentTargetVertexTarget = new FastReadArrayList<>();

            m_fraEdges4TargetIndex.setQuick(targetVertex.getIndex(), edgesWithCurrentTargetVertexTarget);
        }

        // das 'add' hier ist etwas unglücklich, da hier immer nur um 1 erweitert wird...vielleicht doch mit ArrayLists und einer trim()-Methode?
        // Wird dann allerdings sehr schnell recht speicherintensiv...
        edgesWithCurrentTargetVertexTarget.add(newEdge);

    }



    @Override
    public int getBiAdjacencyListSize()
    {
        return m_fraEdges4SourceIndex.size();
    }



    @Override
    public List<E> getEdges()
    {

        int iEdgeCount = 0;

        for (FastReadArrayList<E> edgeArray4CurrentVertex : m_fraEdges4TargetIndex)
            if(edgeArray4CurrentVertex != null) iEdgeCount += edgeArray4CurrentVertex.size();


        FastReadArrayList<E> alEdges = new FastReadArrayList<E>(iEdgeCount);

        int i = 0;
        for (FastReadArrayList<E> edgeArray4CurrentVertex : m_fraEdges4TargetIndex)
        {
            if(edgeArray4CurrentVertex == null) continue;


            for (E edge : edgeArray4CurrentVertex)
                alEdges.setQuick(i++, edge);
        }


        return alEdges;
    }




    /**
     * Gets an edge list. The index of an edges source vertex corresponds to the edges index/position inside this list.
     * 
     * @return An edge list. The index of an edges source vertex corresponds to the edges index/position inside this list.
     */
    public FastReadArrayList<FastReadArrayList<E>> getEdges4SourceIndex()
    {
        return m_fraEdges4SourceIndex;
    }



    /**
     * Gets an edge list. The index of an edges target vertex corresponds to the edges index/position inside this list.
     * 
     * @return An edge list. The index of an edges target vertex corresponds to the edges index/position inside this list.
     */
    public FastReadArrayList<FastReadArrayList<E>> getEdges4TargetIndex()
    {
        return m_fraEdges4TargetIndex;
    }



    @Override
    public List<E> getEdgesWithSource(int vertexIndex)
    {
        FastReadArrayList<E> edgesWithSource = m_fraEdges4SourceIndex.get(vertexIndex);

        if(edgesWithSource != null) return edgesWithSource;


        return new LinkedList<E>();
    }



    @Override
    public List<E> getEdgesWithSource(V vertex)
    {
        return getEdgesWithSource(vertex.getIndex());
    }



    @Override
    public List<E> getEdgesWithTarget(int vertexIndex)
    {
        FastReadArrayList<E> edgesWithTarget = m_fraEdges4TargetIndex.get(vertexIndex);

        if(edgesWithTarget != null) return edgesWithTarget;


        return new LinkedList<E>();
    }



    @Override
    public List<E> getEdgesWithTarget(V vertex)
    {
        return getEdgesWithTarget(vertex.getIndex());
    }



    @Override
    public List<E> getNeighbourEdges(int vertexIndex)
    {
        List<E> edgesWithTarget = getEdgesWithTarget(vertexIndex);
        List<E> edgesWithSource = getEdgesWithSource(vertexIndex);

        FastReadArrayList<E> lNeighbourEdges = new FastReadArrayList<>(edgesWithTarget.size() + edgesWithSource.size());

        lNeighbourEdges.addAll(edgesWithTarget);
        lNeighbourEdges.addAll(edgesWithSource);


        return lNeighbourEdges;
    }



    @Override
    public List<E> getNeighbourEdges(V vertex)
    {
        List<E> edgesWithTarget = getEdgesWithTarget(vertex);
        List<E> edgesWithSource = getEdgesWithSource(vertex);

        FastReadArrayList<E> lNeighbourEdges = new FastReadArrayList<>(edgesWithTarget.size() + edgesWithSource.size());

        lNeighbourEdges.addAll(edgesWithTarget);
        lNeighbourEdges.addAll(edgesWithSource);


        return lNeighbourEdges;
    }



    /**
     * Grows the internal data structures in order to minimize the amount of array copies during addEdge invocations.
     * 
     * @param size the new size of the internal data structures.
     */
    public void grow(int size)
    {
        getEdges4SourceIndex().grow(size);
        getEdges4TargetIndex().grow(size);
    }



    public void setContent(Collection<E> edges)
    {
        this.setContent(edges, false);
    }



    public void setContent(Collection<E> edges, boolean bVerbose)
    {
        m_fraEdges4TargetIndex = new FastReadArrayList<>();
        m_fraEdges4SourceIndex = new FastReadArrayList<>();


        int iCurrentVertexIndex;


        // wir ermitteln die Angaben pro Vertex, wie groß die Edge-Arrays der zweiten Dimension initial allokiert werden müssen. Das erspart relativ
        // teure (besser: sehr häufige) arraycopies

        TIntArrayList edgesWithSourceCountLeftPerVertex = new TIntArrayList();
        TIntArrayList edgesWithTargetCountLeftPerVertex = new TIntArrayList();

        for (E edge : edges)
        {
            iCurrentVertexIndex = edge.getSource().getIndex();

            // wir erweitern das Array falls nötig
            if(iCurrentVertexIndex >= edgesWithSourceCountLeftPerVertex.size())
                edgesWithSourceCountLeftPerVertex.fill(edgesWithSourceCountLeftPerVertex.size(), iCurrentVertexIndex + 1,
                        edgesWithSourceCountLeftPerVertex.getNoEntryValue());

            edgesWithSourceCountLeftPerVertex.setQuick(iCurrentVertexIndex, edgesWithSourceCountLeftPerVertex.get(iCurrentVertexIndex) + 1);



            iCurrentVertexIndex = edge.getTarget().getIndex();

            // wir erweitern das Array falls nötig
            if(iCurrentVertexIndex >= edgesWithTargetCountLeftPerVertex.size())
                edgesWithTargetCountLeftPerVertex.fill(edgesWithTargetCountLeftPerVertex.size(), iCurrentVertexIndex + 1,
                        edgesWithTargetCountLeftPerVertex.getNoEntryValue());

            edgesWithTargetCountLeftPerVertex.setQuick(iCurrentVertexIndex, edgesWithTargetCountLeftPerVertex.get(iCurrentVertexIndex) + 1);

        }


        // nun erstellen wir die Matrix

        int iNodeCount = Math.max(edgesWithSourceCountLeftPerVertex.size(), edgesWithTargetCountLeftPerVertex.size());
        m_fraEdges4SourceIndex.grow(iNodeCount);
        m_fraEdges4TargetIndex.grow(iNodeCount);

        int i = 0;
        for (E edge : edges)
        {
            // die targets für die source
            iCurrentVertexIndex = edge.getSource().getIndex();

            FastReadArrayList<E> edgesWithCurrentVertexSource = m_fraEdges4SourceIndex.get(iCurrentVertexIndex);

            if(edgesWithCurrentVertexSource == null)
            {
                edgesWithCurrentVertexSource = new FastReadArrayList<>(edgesWithSourceCountLeftPerVertex.getQuick(iCurrentVertexIndex));

                m_fraEdges4SourceIndex.setQuick(iCurrentVertexIndex, edgesWithCurrentVertexSource);
            }

            edgesWithCurrentVertexSource.setQuick(edgesWithSourceCountLeftPerVertex.getQuick(iCurrentVertexIndex) - 1, edge);

            edgesWithSourceCountLeftPerVertex.setQuick(iCurrentVertexIndex, edgesWithSourceCountLeftPerVertex.getQuick(iCurrentVertexIndex) - 1);



            // die sources für das target
            iCurrentVertexIndex = edge.getTarget().getIndex();

            FastReadArrayList<E> edgesWithCurrentVertexTarget = m_fraEdges4TargetIndex.get(iCurrentVertexIndex);

            if(edgesWithCurrentVertexTarget == null)
            {
                edgesWithCurrentVertexTarget = new FastReadArrayList<>(edgesWithTargetCountLeftPerVertex.getQuick(iCurrentVertexIndex));

                m_fraEdges4TargetIndex.setQuick(iCurrentVertexIndex, edgesWithCurrentVertexTarget);
            }

            edgesWithCurrentVertexTarget.setQuick(edgesWithTargetCountLeftPerVertex.getQuick(iCurrentVertexIndex) - 1, edge);

            edgesWithTargetCountLeftPerVertex.setQuick(iCurrentVertexIndex, edgesWithTargetCountLeftPerVertex.getQuick(iCurrentVertexIndex) - 1);


            if(bVerbose && ++i % 10000000 == 0) Logger.getLogger(CoreBiAdjacencyList.class.getName()).info("loaded " + StringUtils.beautifyNumber(i) + " edges");
        }
        if(bVerbose) Logger.getLogger(CoreBiAdjacencyList.class.getName()).info("loaded " + StringUtils.beautifyNumber(i) + " edges");


    }






    /**
     * Sets an internal edge list. The index of an edges source vertex corresponds to the edges index/position inside this list.
     * 
     * @param edges4SourceIndex the new internal edge list. The index of an edges source vertex corresponds to the edges index/position inside this list.
     */
    public void setEdges4SourceIndex(FastReadArrayList<FastReadArrayList<E>> edges4SourceIndex)
    {
        m_fraEdges4SourceIndex = edges4SourceIndex;
    }







    /**
     * Sets an internal edge list. The index of an edges target vertex corresponds to the edges index/position inside this list.
     * 
     * @param edges4TargetIndex the new internal edge list. The index of an edges target vertex corresponds to the edges index/position inside this list.
     */
    public void setEdges4TargetIndex(FastReadArrayList<FastReadArrayList<E>> edges4TargetIndex)
    {
        m_fraEdges4TargetIndex = edges4TargetIndex;
    }
}
