package de.dfki.km.jade.graph.core;



import java.io.IOException;
import java.util.Arrays;
import java.util.Collection;
import java.util.List;
import java.util.logging.Logger;

import de.dfki.km.jade.graph.BiAdjacencyList;
import de.dfki.km.jade.graph.BiGraph;
import de.dfki.km.jade.graph.Edge;
import de.dfki.km.jade.graph.Graph;
import de.dfki.km.jade.graph.Vertex;
import de.dfki.km.jade.util.FastReadArrayList;
import de.dfki.km.jade.util.StringUtils;



/**
 * A graph implementation that recognizes node relationships
 * 
 * @param <V> the vertex type
 * @param <E> the edge type
 */
public class CoreBiGraph<V extends Vertex, E extends Edge<V>> implements BiGraph<V, E>
{

    protected CoreBiAdjacencyList<V, E> m_biAdjacencyList;

    protected FastReadArrayList<V> m_vertices = new FastReadArrayList<>();



    public CoreBiGraph()
    {
        m_biAdjacencyList = new CoreBiAdjacencyList<V, E>();
    }



    /**
     * Creates a new {@link CoreBiGraph} Object. The internal index values inside the Vertex Objects will be set according to the position inside the collection.
     * 
     * @param vertices all vertices of the new graph. Also including the orphrans.
     * @param edges all edges of the new graph. Connects the vertex Objects.
     */
    public CoreBiGraph(Collection<V> vertices, Collection<E> edges)
    {
        setContent(vertices, edges);
    }





    /**
     * Creates a new {@link CoreBiGraph} out of the vertices and edges of another graph. The internal index values inside the Vertex Objects will NOT be modified.
     * 
     * @param otherGraph the source graph
     */
    public CoreBiGraph(Graph<V, E> otherGraph)
    {
        m_vertices.addAll(otherGraph.getVertices());

        m_biAdjacencyList = new CoreBiAdjacencyList<V, E>(otherGraph.getEdges());
    }



    /**
     * Creates a new {@link CoreBiGraph} Object. The internal index values inside the Vertex Objects will be set according to the position inside the collection.
     * 
     * @param vertices all vertices of the new graph. Also including the orphrans.
     * @param list a {@link BiAdjacencyList} object with fast edge relationship lookup.
     */
    public CoreBiGraph(V[] vertices, CoreBiAdjacencyList<V, E> list)
    {
        m_vertices.addAll(Arrays.asList(vertices));

        // die vertices kriegen als index ihre Position im Array
        int i = 0;
        for (V vertex : m_vertices)
            vertex.setIndex(i++);

        m_biAdjacencyList = list;
    }



    /**
     * Creates a new {@link CoreBiGraph} Object. The internal index values inside the Vertex Objects will be set according to their positions inside the array.
     * 
     * @param vertices all vertices of the new graph. Also including the orphrans.
     * @param edges all edges of the new graph. Connects the vertex Objects.
     */
    public CoreBiGraph(V[] vertices, E[] edges)
    {
        m_vertices.addAll(Arrays.asList(vertices));

        // die vertices kriegen als index ihre Position im Array
        int i = 0;
        for (V vertex : m_vertices)
            vertex.setIndex(i++);

        m_biAdjacencyList = new CoreBiAdjacencyList<V, E>(Arrays.asList(edges));
    }



    /**
     * Adds an edge to the graph. The method assumes all edge vertices are already added to the graph.<br>
     * <b>CAUTION : </b>This is a very costly method, don't use it to create the whole BiAdjacencyList, only for slight modifications. The internal array structures will
     * be grown on every invocation with arraycopies. This design is in favour of read performance and memory consumtion against incremental write performance. To create
     * the whole BiGraphImpl fast, use the constructors or {@link #setContent(Collection, Collection)}.
     */
    @Override
    public void addEdge(E edge)
    {

        // hmm...also hier kann man komplett über den Graphen traversieren (komplett), indem man einen Knoten hat, und über getSource,-bzw.
        // getTargetEdges sich die Kanten geben lässt, die dann wieder die echten Knoten enthalten. Dazu braucht es keinen lookup über einen
        // EdgeIndex...sprich: einen index in einer Edge braucht man auch nicht. Den kompletten Graphen kriegt man über getVertices() oder getEdges()
        //
        // Bei der GraphImpl verhält es sich etwas anders, hier können wir nicht nach benachbarten Knoten fragen, bzw. nach den Kanten zu einem
        // Knoten. Trotzdem kann man mit getVertices und getEdges den kompletten Graphen kriegen, traversieren wird allerdings etwas...mühsam, was
        // die Sinnhaftigkeit dieser Klasse anzweifeln lässt.
        //
        // Fazit: --Edge.getIndex, --getEdge(index) aus dem Graph und aus dem Entity interface



        // die Knoten sind nun drin - jetzt soll noch die Relation rein. Die isch in der internen BiAjazenzliste
        m_biAdjacencyList.addEdge(edge);

    }







    /**
     * <b>CAUTION : </b>This is a very costly method, don't use it to create the whole BiAdjacencyList, only for slight modifications. The internal array structures will
     * be grow on every invocation with arraycopies. This design is in favour of read performance and memory consumtion against incremental write performance. To create
     * the whole BiGraphImpl fast, use the constructors.
     */
    @Override
    public int addVertex(V vertex)
    {
        m_vertices.add(vertex);

        vertex.setIndex(m_vertices.size() - 1);

        return vertex.getIndex();
    }



    /**
     * This is a REALLY expensive method. If you need it faster, e.g. get all vertices with {@link #getVertices()} once, write them into a HashSet, and check against this
     * structure. We create here the vertices List on the fly, and check inside this. This enables us to save memory in favour of saving runtime, as we didn't need this
     * method for our algorithms.
     */
    @Override
    public boolean contains(E edge)
    {
        return m_biAdjacencyList.getEdges().contains(edge);
    }



    /**
     * Grows the internal data structures in order to minimize the amount of array copies during addEdge and addVertex invocations.
     * 
     * @param newVertexSize
     * @param newEdgeSize
     */
    public void grow(int newVertexSize, int newEdgeSize)
    {
        m_biAdjacencyList.grow(newEdgeSize);
        m_vertices.grow(newVertexSize);
    }



    @Override
    public boolean contains(V vertex)
    {
        if(m_vertices.get(vertex.getIndex()) != null) return true;

        return false;
    }



    public BiAdjacencyList<V, E> getBiAdjacencyList()
    {
        return m_biAdjacencyList;
    }



    @Override
    public int getBiAdjacencyListSize()
    {
        return m_biAdjacencyList.getBiAdjacencyListSize();
    }



    @Override
    public List<E> getEdges()
    {
        return m_biAdjacencyList.getEdges();
    }




    @Override
    public List<E> getEdgesWithSource(int vertexIndex)
    {
        return m_biAdjacencyList.getEdgesWithSource(vertexIndex);
    }



    @Override
    public List<E> getEdgesWithSource(V vertex)
    {
        return m_biAdjacencyList.getEdgesWithSource(vertex);
    }






    @Override
    public List<E> getEdgesWithTarget(int vertexIndex)
    {
        return m_biAdjacencyList.getEdgesWithTarget(vertexIndex);
    }



    @Override
    public List<E> getEdgesWithTarget(V vertex)
    {
        return m_biAdjacencyList.getEdgesWithTarget(vertex);
    }



    @Override
    public List<E> getNeighbourEdges(int vertexIndex)
    {
        List<E> sourceEdges = getEdgesWithTarget(vertexIndex);
        List<E> targetEdges = getEdgesWithSource(vertexIndex);

        FastReadArrayList<E> lNeighbourEdges = new FastReadArrayList<>(sourceEdges.size() + targetEdges.size());

        lNeighbourEdges.addAll(sourceEdges);
        lNeighbourEdges.addAll(targetEdges);


        return lNeighbourEdges;
    }



    @Override
    public List<E> getNeighbourEdges(V vertex)
    {
        List<E> sourceEdges = getEdgesWithTarget(vertex);
        List<E> targetEdges = getEdgesWithSource(vertex);

        FastReadArrayList<E> lNeighbourEdges = new FastReadArrayList<>(sourceEdges.size() + targetEdges.size());

        lNeighbourEdges.addAll(sourceEdges);
        lNeighbourEdges.addAll(targetEdges);


        return lNeighbourEdges;
    }



    @Override
    public V getVertex(int index)
    {
        return m_vertices.get(index);
    }



    @Override
    public List<V> getVertices()
    {
        return m_vertices;
    }





    /**
     * Loads a formerly persisted graph. The format is the one from {@link GraphEntitiesContainer}.
     * 
     * @param strFilePath the graph file, generated by {@link #save(String)}
     * 
     * @throws IOException
     */
    public void load(String strFilePath) throws IOException
    {
        this.load(strFilePath, false);
    }



    /**
     * Loads a formerly persisted graph. The format is the one from {@link GraphEntitiesContainer}.
     * 
     * @param strFilePath the graph file, generated by {@link #save(String)}
     * @param bVerbose true for logs
     * 
     * @throws IOException
     */
    public void load(String strFilePath, boolean bVerbose) throws IOException
    {
        GraphEntitiesContainer<V, E> coreGraph = new GraphEntitiesContainer<>();

        if(bVerbose) Logger.getLogger(CoreBiGraph.class.getName()).info("start reading graph files under " + strFilePath);
        coreGraph.load(strFilePath);

        if(bVerbose)
            Logger.getLogger(CoreBiGraph.class.getName()).info(
                    "will create matrix for " + StringUtils.beautifyNumber(coreGraph.getVertices().size()) + " vertices with "
                            + StringUtils.beautifyNumber(coreGraph.getEdges().size()) + " relationships");

        this.setContentRaw(coreGraph.getVertices(), coreGraph.getEdges(), bVerbose);
    }



    /**
     * Persits this graph into a file as a {@link GraphEntitiesContainer}.
     * 
     * @param strFilePath the target file
     * 
     * @throws IOException
     */
    public void save(String strFilePath) throws IOException
    {
        // wir nehmen zum speichern einen CoreGraphen, da der intern einfach nur 2 Listen hat

        GraphEntitiesContainer<V, E> coreGraph = new GraphEntitiesContainer<>();
        coreGraph.setInternalLists(getVertices(), getEdges());

        coreGraph.save(strFilePath);
    }



    /**
     * Copies the content of the given collections into the internal data structures. The former data will be deleted first. The internal index entries of the vertex
     * Objects will be modified according to their (iterator) position inside the given collections
     * 
     * @param vertices the internal index of these objects will be modified
     * @param edges connects the vertices objects
     */
    public void setContent(Collection<V> vertices, Collection<E> edges)
    {
        m_vertices.clear();

        m_vertices.addAll(vertices);

        // die vertices kriegen als index ihre Position im Array
        int i = 0;
        for (V vertex : m_vertices)
            vertex.setIndex(i++);

        m_biAdjacencyList = new CoreBiAdjacencyList<V, E>(edges);
    }



    /**
     * Copies the content of the given collections into the internal data structures. The former data will be deleted first. The internal index entries of the vertex
     * Objects will <b>NOT</b> be modified.
     * 
     * @param vertices the new, raw vertices
     * @param edges connects the vertices objects
     * @param bVerbose true for logs
     */
    protected void setContentRaw(Collection<V> vertices, Collection<E> edges, boolean bVerbose)
    {
        m_vertices.clear();
        m_vertices.addAll(vertices);

        m_biAdjacencyList = new CoreBiAdjacencyList<V, E>(edges, bVerbose);
    }



}
