package de.dfki.km.jade.graph.core;



import java.util.Arrays;
import java.util.List;

import de.dfki.km.jade.graph.Edge;
import de.dfki.km.jade.graph.Path;
import de.dfki.km.jade.graph.Vertex;
import de.dfki.km.jade.graph.WeightedEntity;
import de.dfki.km.jade.util.FastReadArrayList;



/**
 * A weighted path that is implemented as one-directed linked list. You can go backwards to the 'parent path' with {@link #getParent()}
 * 
 * 
 * @param <V> the vertex type
 * @param <E> the edge type
 */
public class CorePath<V extends Vertex, E extends Edge<V>> implements Path<V, E>, WeightedEntity, Comparable<CorePath<V, E>>
{
    protected E m_endEdge;

    protected V m_endVertex;

    protected int m_length;

    protected CorePath<V, E> m_parent;

    protected V m_startVertex;

    protected float m_weight = 0f;



    public CorePath(V startAndEndVertex)
    {
        m_length = 1;

        m_endEdge = null;

        m_endVertex = startAndEndVertex;

        m_startVertex = startAndEndVertex;

        m_parent = null;

    }




    public CorePath(V endVertex, E endEdge, CorePath<V, E> parent)
    {
        m_endEdge = endEdge;

        m_endVertex = endVertex;

        m_parent = parent;

        m_startVertex = parent.getStartVertex();

        m_length = m_parent.getLength() + 1;
    }





    @Override
    public void addEdge(E edge)
    {
        // TODO soll die methode überhaupt im Interface sein?
        throw new UnsupportedOperationException("CorePaths can not be modified");
    }






    @Override
    public int addVertex(V vertex)
    {
        // TODO soll die methode überhaupt im Interface sein?
        throw new UnsupportedOperationException("CorePaths can not be modified");
    }



    @Override
    public void addWeight(float weight2add)
    {
        m_weight += weight2add;
    }



    /**
     * Appends a corepath to this one. Returns a new path Object with the result. Note that the paths must have the same end vertex in order to merge
     * them. The current Object (this) will not be modified.
     * 
     * @param corePath2append the corePath to append
     * 
     * @return the merge result
     */
    public CorePath<V, E> append(CorePath<V, E> corePath2append)
    {
        if(!this.getEndVertex().equals(corePath2append.getEndVertex()))
            throw new IllegalStateException("CorePaths must have the same end vertex in order to merge them");


        CorePath<V, E> newPath = this;

        CorePath<V, E> restPath2append = corePath2append;

        while (restPath2append.getEndEdge() != null)
        {
            V newEndVertex = restPath2append.getEndEdge().getOtherVertex(restPath2append.getEndVertex());

            float newWeight = this.getWeight() + corePath2append.getWeight() - restPath2append.getParent().getWeight();

            newPath = new CorePath<>(newEndVertex, restPath2append.getEndEdge(), newPath);
            newPath.setWeight(newWeight);

            restPath2append = restPath2append.getParent();
        }



        return newPath;
    }





    @Override
    public int compareTo(CorePath<V, E> other)
    {
        if(m_weight < other.getWeight()) return -1;
        if(m_weight > other.getWeight()) return 1;


        return 0;
    }



    @Override
    public boolean contains(E edge)
    {
        if(getEndEdge() == edge) return true;

        if(getEndEdge().equals(edge)) return true;

        if(m_parent != null) return m_parent.contains(edge);


        return false;
    }






    @Override
    public boolean contains(V vertex)
    {
        if(getEndVertex().getIndex() == vertex.getIndex()) return true;

        if(m_parent != null) return m_parent.contains(vertex);


        return false;
    }



    @SuppressWarnings("unchecked")
    public boolean equals(Object o)
    {
        CorePath<V, E> other = (CorePath<V, E>) o;

        if(m_length != other.getLength()) return false;

        if(m_endEdge != other.getEndEdge()) return false;

        if(m_parent == null) return true;



        return m_parent.equals(other.getParent());
    }



    public E getEdge(int indexOfEdgeInPath)
    {

        // getLength ist die Anzahl der Vertices - Edges gibt es eines weniger. Und dann noch eins weg, um von der size auf den index zu kommen (ab
        // 0)
        int iCurrentEdgeIndex = getLength() - 2;
        if(indexOfEdgeInPath > iCurrentEdgeIndex)
            throw new ArrayIndexOutOfBoundsException("there is no edge in this path with index " + indexOfEdgeInPath);


        CorePath<V, E> currentPathCursor = this;

        while (indexOfEdgeInPath != iCurrentEdgeIndex)
        {
            currentPathCursor = currentPathCursor.getParent();

            indexOfEdgeInPath--;
        }


        return currentPathCursor.getEndEdge();
    }



    /**
     * Note that the List will be generated on the fly, backed by an Array. Don't use this method extensively, reuse the return value.
     */
    @Override
    public List<E> getEdges()
    {
        FastReadArrayList<E> edges = new FastReadArrayList<>(getLength() - 1);

        CorePath<V, E> currentPathCursor = this;

        // im ersten PathCursor gibt es keine Edge (startNode==endNode)
        for (int i = getLength() - 1; i > 0; i--)
        {
            edges.setQuick(i, currentPathCursor.getEndEdge());

            currentPathCursor = currentPathCursor.getParent();
        }


        return edges;
    }



    @Override
    public E getEndEdge()
    {
        return m_endEdge;
    }



    @Override
    public V getEndVertex()
    {
        return m_endVertex;
    }



    @Override
    public List<V> getLeafs()
    {
        return Arrays.asList(m_endVertex);
    }



    public final int getLength()
    {
        return m_length;
    }



    /**
     * Gives the 'parent' path, which is the same path as the current one, but without the last hop (one shorter).
     * 
     * @return the 'parent' path, which is the same path as the current one, but without the last hop (one shorter).
     */
    public CorePath<V, E> getParent()
    {
        return m_parent;
    }



    @Override
    public V getRoot()
    {
        return m_startVertex;
    }



    @Override
    public E getStartEdge()
    {
        CorePath<V, E> currentPathCursor = this;
        E firstEdge = null;

        while (true)
        {
            if(currentPathCursor.getEndEdge() == null) return firstEdge;

            firstEdge = currentPathCursor.getEndEdge();

            // wenn es eine edge gibt, dann gibt es auch einen parent
            currentPathCursor = currentPathCursor.getParent();
        }

    }





    @Override
    public V getStartVertex()
    {
        return m_startVertex;
    }





    /**
     * Note that this is not the index/position of the vertex inside the path, but the nodeID / vertexID the vertex object has inside the whole,
     * underlying graph. For fast random access, Invoke {@link #getVertices()} ONCE and use the returned List.
     */
    @Override
    public V getVertex(int index)
    {

        CorePath<V, E> currentPathCursor = this;

        while (true)
        {
            V currentVertex = currentPathCursor.getEndVertex();
            if(currentVertex.getIndex() == index) return currentVertex;

            currentPathCursor = currentPathCursor.getParent();

            if(currentPathCursor == null) return null;
        }

    }



    @Override
    public List<V> getVertices()
    {
        FastReadArrayList<V> vertices = new FastReadArrayList<>(getLength());

        CorePath<V, E> currentPathCursor = this;

        for (int i = getLength() - 1; i >= 0; i--)
        {
            vertices.setQuick(i, currentPathCursor.getEndVertex());

            currentPathCursor = currentPathCursor.getParent();
        }


        return vertices;
    }



    @Override
    public float getWeight()
    {
        return m_weight;
    }



    @Override
    public boolean isEdgeDirectionForward(int indexOfEdgeInPath)
    {

        // getLength ist die Anzahl der Vertices - Edges gibt es eines weniger. Und dann noch eins weg, um von der size auf den index zu kommen (ab
        // 0)
        int iCurrentEdgeIndex = getLength() - 2;
        if(indexOfEdgeInPath > iCurrentEdgeIndex)
            throw new ArrayIndexOutOfBoundsException("there is no edge in this path with index " + indexOfEdgeInPath);


        CorePath<V, E> currentPathCursor = this;

        while (indexOfEdgeInPath != iCurrentEdgeIndex)
        {
            currentPathCursor = currentPathCursor.getParent();

            indexOfEdgeInPath--;
        }


        return currentPathCursor.isEndEdgeDirectionForward();
    }







    /**
     * Gives the information if the last edge, as part of this path, goes from left to right (forward), where the direction goes from source node to
     * target node - or vice versa. You can also get this information by looking to the source/target vertex and look to the last edge inside this
     * path.
     * 
     * @return true if the edge is a forward-edge, thus it goes from left to right (source=>target). False otherwise.
     */
    public Boolean isEndEdgeDirectionForward()
    {
        return (m_endEdge != null && m_endEdge.getTarget().equals(m_endVertex));
    }





    @Override
    public void setWeight(float weight)
    {
        m_weight = weight;
    }



    @Override
    public void subtractWeight(float weight2subtract)
    {
        m_weight -= weight2subtract;
    }



    public String toString()
    {
        StringBuilder sbPath = new StringBuilder();


        CorePath<V, E> currentCursor = this;

        for (int i = 0; i < m_length; i++)
        {
            if(currentCursor.getEndEdge() == null) continue;


            sbPath.insert(0, currentCursor.getEndVertex().toString());

            if(currentCursor.isEndEdgeDirectionForward())
                sbPath.insert(0, ">>");
            else
                sbPath.insert(0, "<<");


            currentCursor = currentCursor.getParent();
        }
        sbPath.insert(0, currentCursor.getStartVertex());


        sbPath.append("_w");
        sbPath.append(this.getWeight());


        return sbPath.toString();
    }
}
