package de.dfki.km.jade.graph.core;



import java.util.Collection;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Set;

import de.dfki.km.jade.graph.Edge;
import de.dfki.km.jade.graph.Vertex;
import de.dfki.km.jade.graph.search.SubGraph;
import de.dfki.km.jade.util.FastReadArrayList;



/**
 * The core {@link SubGraph} implementation.<br>
 * <b>Note</b>: this comparator imposes orderings that are inconsistent with equals.
 * 
 * 
 * @param <V> the vertex type
 * @param <E> the edge type
 */
public class CoreSubGraph<V extends Vertex, E extends Edge<V>> implements SubGraph<V, E>, Comparable<CoreSubGraph<V, E>>
{
    @SuppressWarnings("rawtypes")
    static protected class StartVertexComparator implements Comparator<CorePath>
    {

        @Override
        public int compare(CorePath o1, CorePath o2)
        {
            return Integer.compare(o1.getStartVertex().getIndex(), o2.getStartVertex().getIndex());
        }

    }


    protected static StartVertexComparator startVertexComparator = new StartVertexComparator();



    @SuppressWarnings("rawtypes")
    protected static <V extends Vertex, E extends Edge<V>> int getPathsLengthSum(SubGraph<V, E> subGraph)
    {
        int iCorePathLenghtSum = 0;
        for (CorePath path : subGraph.getPaths())
            iCorePathLenghtSum += path.getLength();

        return iCorePathLenghtSum;
    }


    protected boolean m_bDifferentRootIsEqual = false;

    protected float m_weight;

    protected Set<E> m_hsEdges;

    protected List<CorePath<V, E>> m_lCorePaths;

    protected List<V> m_lLeafs;

    protected V m_rootVertex;



    public CoreSubGraph(V root)
    {
        m_rootVertex = root;

        m_weight = 0.0f;

        m_hsEdges = new HashSet<>();

        m_lLeafs = new LinkedList<>();

        m_lCorePaths = new LinkedList<CorePath<V, E>>();

    }




    @SuppressWarnings("unchecked")
    @Override
    public void addEdge(E edge)
    {
        CorePath<V, E> path = (CorePath<V, E>) new CorePath<>(edge.getTarget(), edge, new CorePath<>(edge.getSource()));

        addPath(path);
    }







    @Override
    public void addGraphPaths(SubGraph<V, E> subGraph)
    {
        addPaths(subGraph.getPaths());
    }



    @Override
    public void addPath(CorePath<V, E> corePath)
    {
        CorePath<V, E> currentCorePath = corePath;

        m_weight = m_weight + corePath.getWeight();

        while (currentCorePath != null)
        {
            if(currentCorePath.getEndEdge() != null)
            {
                m_hsEdges.add(currentCorePath.getEndEdge());

                currentCorePath = currentCorePath.getParent();
            }

            else
            {
                break;
            }
        }




        // ist das performancekritisch, wenn das eine SortedList ist? Die tests sagen bis jetzt nein:) Ich brauche eine einheitliche Reihenfolge der
        // internen Pfade, wenn mit getVertices die Knotenreihenfolge auch einheitlich sein soll. Nun ist diese bei Pfaden mal von rechts nach links,
        // mal von links nach rechts. Da kann ich dann nicht mehr mit dem Levenshtein testen
        int iIndex = Collections.binarySearch(m_lCorePaths, corePath, startVertexComparator);
        if(iIndex < 0) iIndex = (iIndex + 1) * -1;

        m_lCorePaths.add(iIndex, corePath);

        m_lLeafs.add(iIndex, corePath.getStartVertex());
    }



    @Override
    public void addPaths(Collection<CorePath<V, E>> paths)
    {
        for (CorePath<V, E> cursor : paths)
            addPath(cursor);
    }



    /**
     * Adds a vertex to this subgraph. Because this is a subgraph, the internal index of vertex will not be modified. The method assumes that it is
     * correctly set as part of another, whole 'parent' graph.
     * 
     * @param vertex the vertex that should be added. The internal index will NOT be modified, because this is a subgraph, and the vertex is already
     *            part of a 'bigger' graph structure.
     * 
     * @return the UNMODIFIED internal index of vertex
     */
    @Override
    public int addVertex(V vertex)
    {
        addPath(new CorePath<V, E>(vertex));

        return vertex.getIndex();
    }



    @Override
    public void addWeight(float weight2add)
    {
        m_weight += weight2add;
    }




    @Override
    public int compareTo(CoreSubGraph<V, E> otherSubGraph)
    {
        if(m_weight < otherSubGraph.getWeight()) return -1;
        if(m_weight > otherSubGraph.getWeight()) return 1;
        
        
        return 0;
    }



    @Override
    public boolean contains(E edge)
    {
        return m_hsEdges.contains(edge);
    }







    @Override
    public boolean contains(V vertex)
    {
        for (CorePath<V, E> corePath : m_lCorePaths)
        {
            if(corePath.contains(vertex)) return true;
        }


        return false;
    }



    @SuppressWarnings("unchecked")
    @Override
    public boolean equals(Object obj)
    {
        SubGraph<V, E> casted = (SubGraph<V, E>) obj;

        if(m_bDifferentRootIsEqual) return equalsWithoutRoot(casted);


        return super.equals(obj);
    }



    protected boolean equalsWithoutRoot(SubGraph<V, E> otherGraph)
    {

        if(this == otherGraph) return true;

        if(m_hsEdges.size() != otherGraph.getEdges().size()) return false;
        if(m_hsEdges.size() == 0) return true;

        // die mEdges-Länge ist nicht gleich der Pfadlänge - u.A. werden Pfade mit single-gegenüber doppeleinträgen bei Edges gleich bewertet (da Set)
        int iCorePathLenghtSumThis = getPathsLengthSum(this);
        int iCorePathLenghtSumOther = getPathsLengthSum(otherGraph);

        if(iCorePathLenghtSumThis != iCorePathLenghtSumOther) return false;


        // die Längen haben wir - nun die inhaltliche Gleichheit
        if(!m_hsEdges.containsAll(otherGraph.getEdges())) return false;


        return true;
    }





    @Override
    public List<CorePath<V, E>> getPaths()
    {
        return m_lCorePaths;
    }







    @Override
    public List<E> getEdges()
    {
        return new FastReadArrayList<>(m_hsEdges);
    }



    @Override
    public List<V> getLeafs()
    {
        return m_lLeafs;
    }



    @Override
    public V getRoot()
    {
        return m_rootVertex;
    }




    @Override
    public V getVertex(int index)
    {

        for (CorePath<V, E> corePath : m_lCorePaths)
        {
            V vertex = corePath.getVertex(index);

            if(vertex != null) return vertex;
        }


        return null;
    }



    /**
     * Gets all vertices of all internal paths. The root node will appear only once, other nodes can apear more often in the case they appear multiple
     * times inside the paths. Per definition, the root node is the common node of all subGraphs.
     */
    @Override
    public List<V> getVertices()
    {

        int iOverallLength = 0;

        for (CorePath<V, E> corePath : m_lCorePaths)
            iOverallLength += corePath.getLength();

        // size calculation: rootNode only once
        FastReadArrayList<V> lVertices = new FastReadArrayList<>(iOverallLength - m_lCorePaths.size() + 1);

        int i = 0;
        int iPathIndex = 0;
        for (CorePath<V, E> corePath : m_lCorePaths)
        {
            for (V vertex : corePath.getVertices())
            {
                // we insert this in this way so that subGraphs that have two internal paths will be concatenated in right vertex direction, so you
                // can see easily the path and also deal with the vertex positions
                if(iPathIndex == 0)
                {
                    lVertices.setQuick(i, vertex);
                }
                else
                {
                    if(vertex.equals(m_rootVertex)) continue;

                    int iIndexFromBehind = iOverallLength - 1 - i + m_lCorePaths.get(0).getLength() - 1;

                    lVertices.setQuick(iIndexFromBehind, vertex);
                }

                i++;
            }
            iPathIndex++;
        }

        return lVertices;
    }





    @Override
    public float getWeight()
    {
        return m_weight;
    }



    @Override
    public int hashCode()
    {
        if(m_bDifferentRootIsEqual) return hashCodeWithoutRoot();


        return super.hashCode();
    }



    protected int hashCodeWithoutRoot()
    {
        // nur die mEdges reichen nicht aus, da eine edge auch durchaus 2 mal in einem Pfad sein kann, mEdges aber diese Info als HashSet verliert
        int iHashCode = m_hsEdges.hashCode();

        // aus AbstractList.hashCode()
        iHashCode = 31 * iHashCode + getPathsLengthSum(this);

        return iHashCode;
    }



    public void setResultEqualityWithoutRoots(boolean differentRootIsEqual)
    {
        this.m_bDifferentRootIsEqual = differentRootIsEqual;
    }




    @Override
    public void setWeight(float weight)
    {
        m_weight = weight;
    }





    @Override
    public void subtractWeight(float weight2subtract)
    {
        m_weight -= weight2subtract;
    }






    public String toString()
    {
        final StringBuilder sbSubGraph = new StringBuilder();

        sbSubGraph.append("Root:").append(getRoot().toString());
        sbSubGraph.append(", Leafs:").append(getLeafs().toString());
        sbSubGraph.append(", Weight:").append(getWeight());
        sbSubGraph.append(", Paths:").append(getPaths().toString());


        return sbSubGraph.toString();
    }
}
