package de.dfki.km.jade.graph.core;



import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.List;

import com.esotericsoftware.kryo.Kryo;
import com.esotericsoftware.kryo.io.Input;
import com.esotericsoftware.kryo.io.Output;

import de.dfki.km.jade.graph.Edge;
import de.dfki.km.jade.graph.Graph;
import de.dfki.km.jade.graph.Vertex;



/**
 * The simplest graph implementation, done by an edge and a vertices list. Can be used to create a new graph in a performant way, e.g. by incrementally adding edges. Can
 * be converted relatively fast to a {@link CoreBiGraph}.
 * 
 * 
 * @param <V> the vertex type
 * @param <E> the edge type
 */
public class GraphEntitiesContainer<V extends Vertex, E extends Edge<V>> implements Graph<V, E>
{


    protected List<E> m_edges = new ArrayList<E>();

    protected List<V> m_vertices = new ArrayList<V>();



    public GraphEntitiesContainer()
    {
    }




    /**
     * Creates a new {@link GraphEntitiesContainer} object. the parameter content will be copied into the internal data structures. In the case you want to set the
     * internal lists explicitely, use {@link #setInternalLists(List, List)}
     * 
     * @param vertices the vertices
     * @param edges the edges
     */
    public GraphEntitiesContainer(Collection<V> vertices, Collection<E> edges)
    {
        m_edges.addAll(edges);
        m_vertices.addAll(vertices);
    }



    public GraphEntitiesContainer(V[] vertices, E[] edges)
    {
        m_edges.addAll(Arrays.asList(edges));
        m_vertices.addAll(Arrays.asList(vertices));
    }



    @Override
    public void addEdge(E edge)
    {
        m_edges.add(edge);
    }



    @Override
    public int addVertex(V vertex)
    {
        m_vertices.add(vertex);

        vertex.setIndex(m_vertices.size() - 1);

        return vertex.getIndex();
    }



    @Override
    public boolean contains(E edge)
    {
        return m_edges.contains(edge);
    }




    //
    // @Override
    // public Graph duplicate()
    // {
    // Edge edge;
    //
    // Vertex clone;
    //
    // Vertex origin;
    //
    // final Vertex[] vertices = new Vertex[mVertices.length];
    //
    // final HashMap<Integer, Vertex> map = new HashMap<Integer, Vertex>();
    //
    // for (int i = 0; i < mVertices.length; i++)
    // {
    // origin = mVertices[i];
    //
    // clone = origin.duplicate();
    //
    // vertices[i] = clone;
    //
    // map.put(origin.getIndex(), clone);
    // }
    //
    // final Edge[] edges = new Edge[mEdges.length];
    //
    // for (int i = 0; i < mEdges.length; i++)
    // {
    // edge = mEdges[i];
    //
    // edges[i] = edge.duplicate(map.get(edge.getSource().getIndex()), map.get(edge.getTarget().getIndex()));
    // }
    //
    // final GraphImpl toReturn = new GraphImpl(mIndex, vertices, edges);
    //
    //
    // toReturn.mDirected = mDirected;
    //
    //
    // return toReturn;
    // }



    @Override
    public boolean contains(V vertex)
    {
        if(m_vertices.get(vertex.getIndex()) != null) return true;

        return false;
    }



    @Override
    public boolean equals(Object obj)
    {
        if(this == obj) return true;
        if(obj == null) return false;
        if(getClass() != obj.getClass()) return false;
        @SuppressWarnings("rawtypes")
        GraphEntitiesContainer other = (GraphEntitiesContainer) obj;
        if(!m_edges.equals(other.m_edges)) return false;
        if(!m_vertices.equals(other.m_vertices)) return false;
        return true;
    }





    public List<E> getEdges()
    {
        return m_edges;
    }



    @Override
    public V getVertex(int index)
    {
        return m_vertices.get(index);
    }



    public List<V> getVertices()
    {
        return m_vertices;
    }






    @Override
    public int hashCode()
    {
        final int prime = 31;
        int result = 1;
        result = prime * result + m_edges.hashCode();
        result = prime * result + m_vertices.hashCode();

        return result;
    }



    /**
     * Loads a formerly persisted graph
     * 
     * @param strFilePath the graph file, generated by {@link #save(String)}
     * 
     * @throws IOException
     */
    @SuppressWarnings("unchecked")
    public void load(String strFilePath) throws IOException
    {

        // mit GSON funzt das mit den mehrfach genutzten Objektreferenzen nicht:

        // Type graphType = new TypeToken<GraphEntitiesContainer<V, E>>()
        // {
        // }.getType();
        //
        //
        // JsonReader jsonReader = new JsonReader(new InputStreamReader(new FileInputStream(strFilePath), "UTF-8"));
        //
        //
        // GraphEntitiesContainer<V, E> coreGraph = new Gson().fromJson(jsonReader, graphType);
        //
        // jsonReader.close();
        //
        // this.setInternalLists(coreGraph.getVertices(), coreGraph.getEdges());


        // Versuch mit kryo:
        // für riesige Datenmengen, wenn es mit dem Speicher beim serialisieren eng wird, kann man es auch so probieren:
        // Kryo kryo = new Kryo(new MapDBReferenceResolver());
        Kryo kryo = new Kryo();

        // das ist von einem speziellen kryo-branch, der gut für deep graphes sein soll. Wir sind mit dem Original auf die Nase gefallen..und mit dem auch :(
        kryo.setSupportsContinuations(true);

        Input input = new Input(new FileInputStream(strFilePath));
        GraphEntitiesContainer<V, E> container = kryo.readObject(input, GraphEntitiesContainer.class);
        input.close();


        this.setInternalLists(container.getVertices(), container.getEdges());
    }





    /**
     * Persits this graph into a file
     * 
     * @param strFilePath thhe file where to persist
     * 
     * @throws IOException
     */
    public void save(String strFilePath) throws IOException
    {

        // mit GSON funzt das mit den Objektreferenzen nicht, da die Ihre dafür gedachte Klasse GraphAdapterBuilder nicht an die aktuellen
        // Schnittstellen angepasst haben...und out of the box ist es nicht unterstützt. Somit bekommt man beim deserialisieren n verschiedene
        // Objekte, wo vorher genau eins war. Das geht gar nicht.
        // Type graphType = new TypeToken<GraphEntitiesContainer<V, E>>()
        // {
        // }.getType();
        //
        //
        // OutputStreamWriter outputStreamWriter = new OutputStreamWriter(new FileOutputStream(strFilePath), "UTF-8");
        // JsonWriter writer = new JsonWriter(new BufferedWriter(outputStreamWriter));
        //
        //
        // new Gson().toJson(this, graphType, writer);
        //
        // writer.close();


        // wir probieren das jetzt mal mit kryo - das sieht auf den ersten Blick mal ganz schick aus
        // für riesige Datenmengen, wenn es mit dem Speicher beim serialisieren eng wird, kann man es auch so probieren:
        // Kryo kryo = new Kryo(new MapDBReferenceResolver());
        Kryo kryo = new Kryo();

        // das ist von einem speziellen kryo-branch, der gut für deep graphes sein soll. Wir sind mit dem Original auf die Nase gefallen..und mit dem auch :(
        kryo.setSupportsContinuations(true);

        Output output = new Output(new FileOutputStream(strFilePath));

        kryo.writeObject(output, this);
        output.close();

    }





    /**
     * Sets the internal content lists - no content copies. This is to make quickly some vertex/edge containers. The content of the lists will NOT modified, thus, also
     * the internal vertex indices stay untouched. This differs from the {@link #addVertex(Vertex)} and the {@link CoreBiGraph#setContent(Collection, Collection)}
     * methods.
     * 
     * @param vertices the new internal vertex list
     * @param edges the new internal edge list
     */
    public void setInternalLists(List<V> vertices, List<E> edges)
    {
        m_vertices = vertices;
        m_edges = edges;
    }
}
