package de.dfki.km.jade.graph.core;



import de.dfki.km.jade.graph.Entity;
import de.dfki.km.jade.graph.Vertex;
import de.dfki.km.jade.graph.WeightedEntity;



/**
 * Has a weight, and wrapps entities. The class uses the hashCode and equals methods from the wrapped entity, the comparable interface will be
 * implemented according the weight, equals-objects (according the wrapped entities) will also return 0 in {@link #compareTo(WeightedEntity)}
 * 
 * 
 * @author reuschling
 * 
 * @param <K>
 */
public class WeightedEntityContainer<K extends Entity> implements Entity, WeightedEntity, Comparable<WeightedEntity>
{

    protected K m_entity;

    protected float m_weight = 0;



    public WeightedEntityContainer(K entity)
    {
        m_entity = entity;
    }



    public WeightedEntityContainer(K entity, float weight)
    {
        m_entity = entity;
        m_weight = weight;
    }



    @Override
    public void addWeight(float weight2add)
    {
        m_weight += weight2add;
    }



    @SuppressWarnings("rawtypes")
    @Override
    public int compareTo(WeightedEntity o)
    {

        
        // gleiches Gewicht für nur dazu, daß sie anders sortiert werden
        if(m_weight < o.getWeight()) return -1;
        if(m_weight > o.getWeight()) return 1;


        // jetzt gehts um die Entities
        Entity entityThis = this.getEntity();
        Entity entityOther = (Entity) ((o instanceof WeightedEntityContainer) ? ((WeightedEntityContainer) o).getEntity() : o);

        // System.identityHashCode is not fast as a rocket....this is to boost Vertex performance
        if(entityThis instanceof Vertex && entityOther instanceof Vertex)
        {
            Vertex entityThisAsVertex = (Vertex) entityThis;
            Vertex entityOtherAsVertex = (Vertex) entityOther;

            return Integer.compare(entityThisAsVertex.getIndex(), entityOtherAsVertex.getIndex());
        }
        
        // gleiche entities werden als duppletten aus einem Set entfernt
        if(this.equals(o)) return 0;



        return Integer.compare(System.identityHashCode(entityThis), System.identityHashCode(entityOther));
    }



    @SuppressWarnings("unchecked")
    @Override
    public boolean equals(Object obj)
    {
        if(obj instanceof WeightedEntityContainer)
            return m_entity.equals(((WeightedEntityContainer<K>) obj).getEntity());
        else
            return m_entity.equals(obj);
    }



    public K getEntity()
    {
        return m_entity;
    }



    @Override
    public float getWeight()
    {
        return m_weight;
    }



    @Override
    public int hashCode()
    {
        return m_entity.hashCode();
    }



    public void setEntity(K entity)
    {
        this.m_entity = entity;
    }



    @Override
    public void setWeight(float weight)
    {
        m_weight = weight;
    }



    @Override
    public void subtractWeight(float weight2subtract)
    {
        m_weight -= weight2subtract;
    }



    @Override
    public String toString()
    {
        return "[weight:" + m_weight + " entity: " + m_entity.toString() + "]";
    }





}
