package de.dfki.km.jade.graph.multisteinertree;



import java.util.Collections;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Set;
import java.util.SortedSet;
import java.util.TreeSet;

import de.dfki.km.jade.graph.BiGraph;
import de.dfki.km.jade.graph.Edge;
import de.dfki.km.jade.graph.Vertex;
import de.dfki.km.jade.graph.core.CorePath;
import de.dfki.km.jade.graph.core.CoreSubGraph;
import de.dfki.km.jade.graph.search.EntitySearch;
import de.dfki.km.jade.graph.search.SubGraph;
import de.dfki.km.jade.graph.search.WalkedPathsInterruptConfig;
import de.dfki.km.jade.graph.search.WeightCalculator;
import de.dfki.km.jade.graph.search.WeightedPathExpander;
import de.dfki.km.jade.util.SaveNonEqualSetObjectsComparator;



/**
 * An {@link EntitySearch} implementation, searching for subgraphs that connects given sets of start nodes.
 * 
 * 
 * @param <V> the vertex type
 * @param <E> the edge type
 */
public class MultiSteinerTree<V extends Vertex, E extends Edge<V>> implements EntitySearch<V, E, SubGraph<V, E>>
{
    protected boolean m_bDifferentRootIsEqual = false;

    protected WeightedPathExpander<V, E> m_currentUsedPathExpander;

    protected BiGraph<V, E> m_graph;

    protected WalkedPathsInterruptConfig m_interruptConfig;

    protected int m_iWalkedPathCount = 0;

    protected LinkedList<WeightedPathExpander<V, E>> m_llPathExpanders4StartNodeSets;

    protected TreeSet<SubGraph<V, E>> m_tsSortedResults;

    protected WeightCalculator<V, E, SubGraph<V, E>> m_weightCalculator;





    public MultiSteinerTree(BiGraph<V, E> graph, WeightCalculator<V, E, SubGraph<V, E>> weightCalculator, WalkedPathsInterruptConfig interruptConfig)
    {
        m_graph = graph;

        m_weightCalculator = weightCalculator;

        m_interruptConfig = interruptConfig;
    }



    protected boolean allPathExpandersFoundPathsWithEnd(V endVertex)
    {
        for (WeightedPathExpander<V, E> pathExpander : m_llPathExpanders4StartNodeSets)
        {
            if(pathExpander.getCheapestPathsFoundWithEnd(endVertex) == null) return false;
        }

        return true;
    }



    /**
     * Combines lists of paths with a path list. The entries will be combined pairwise, each resulting combination list comprises an entry of list one
     * (which is a list of paths) and an entry of list two (which is a single path). All combinations will be returned themself as List.
     * 
     * @param lListsOfPaths lists of paths that should be combined pairwise with the given path list
     * @param lPaths a single list of paths, the entries should be combined with the lists of paths
     * 
     * @return the result combination list. Each entry comprises an entry of list one (which is a list of paths) and an entry of list two (which is a
     *         single path)
     */
    protected List<List<CorePath<V, E>>> combinePathLists(List<List<CorePath<V, E>>> lListsOfPaths, List<CorePath<V, E>> lPaths)
    {

        List<List<CorePath<V, E>>> llPathCombinations = new LinkedList<>();


        for (List<CorePath<V, E>> lInitialPathCombination : lListsOfPaths)
        {
            for (CorePath<V, E> path4furtherCombinations : lPaths)
            {
                LinkedList<CorePath<V, E>> llNewCombi = new LinkedList<>();

                llNewCombi.addAll(lInitialPathCombination);
                llNewCombi.add(path4furtherCombinations);

                llPathCombinations.add(llNewCombi);
            }
        }


        return llPathCombinations;
    }



    protected void combineWithOtherPathExpandersPaths(CorePath<V, E> cheapestPathFound)
    {
        // der neu gefundene Pfad wird mit allen bereits gefundenen Pfaden der anderen PathExpanders mit dem selben end nodes kombiniert. Das passiert
        // nur, wenn alle anderen - sofern alle anderen pathExpanders auch schon Pfade für diesen node gefunden haben. Ansonsten haben wir mit diesem
        // Pfad (noch) kein Endergebnis

        V root4NewResultSubGraph = cheapestPathFound.getEndVertex();

        // haben alle einen entsprechenden Pfad?
        if(!allPathExpandersFoundPathsWithEnd(root4NewResultSubGraph)) return;



        List<List<CorePath<V, E>>> lPathCombinations = new LinkedList<>();
        lPathCombinations.add(Collections.singletonList(cheapestPathFound));

        for (WeightedPathExpander<V, E> pathExpander : m_llPathExpanders4StartNodeSets)
        {
            if(pathExpander == m_currentUsedPathExpander) continue;

            lPathCombinations = combinePathLists(lPathCombinations, pathExpander.getCheapestPathsFoundWithEnd(root4NewResultSubGraph));
        }




        for (List<CorePath<V, E>> lPathCombination : lPathCombinations)
        {
            // potential result
            CoreSubGraph<V, E> subGraph4Combination = new CoreSubGraph<>(root4NewResultSubGraph);
            subGraph4Combination.setResultEqualityWithoutRoots(m_bDifferentRootIsEqual);

            subGraph4Combination.addPaths(lPathCombination);


            if(!m_weightCalculator.filterResult(subGraph4Combination, m_tsSortedResults))
            {
                m_tsSortedResults.add(subGraph4Combination);
            }
        }

    }



    public int getCurrentNumberOfExpandedPaths()
    {
        int iNumberOfExpandedPaths = 0;

        for (WeightedPathExpander<V, E> pathExpander : m_llPathExpanders4StartNodeSets)
            iNumberOfExpandedPaths += pathExpander.getNumberOfWalkedPaths();


        return iNumberOfExpandedPaths;
    }



    @Override
    public int getCurrentNumberOfWalkedPaths()
    {
        return m_iWalkedPathCount;
    }



    @Override
    public int getCurrentResultCount()
    {
        return m_tsSortedResults.size();
    }




    @Override
    public TreeSet<SubGraph<V, E>> getCurrentResults()
    {
        return m_tsSortedResults;
    }



    @Override
    public BiGraph<V, E> getGraph()
    {
        return m_graph;
    }



    /**
     * Searches for subgraphs connecting given start node sets.
     */
    public TreeSet<SubGraph<V, E>> search(List<Set<V>> startNodeSets)
    {

        m_tsSortedResults = new TreeSet<SubGraph<V, E>>(new SaveNonEqualSetObjectsComparator<SubGraph<V, E>>());
        m_llPathExpanders4StartNodeSets = new LinkedList<>();


        if(startNodeSets.size() == 0) return m_tsSortedResults;


        for (Set<V> startNodeSet : startNodeSets)
            m_llPathExpanders4StartNodeSets.add(new WeightedPathExpander<>(startNodeSet, m_graph, m_weightCalculator));



        m_interruptConfig.startSearch();


        LinkedList<WeightedPathExpander<V, E>> llNonFinishedPathExpanders = new LinkedList<>(m_llPathExpanders4StartNodeSets);

        boolean bExpandedPathsOverLimit = false;

        wholeSearchLoop: while (llNonFinishedPathExpanders.size() > 0)
        {

            Iterator<WeightedPathExpander<V, E>> itNonFinishedPathExpanders = llNonFinishedPathExpanders.iterator();
            while (itNonFinishedPathExpanders.hasNext())
            {
                m_currentUsedPathExpander = itNonFinishedPathExpanders.next();


                if(m_interruptConfig.interrupt(this)) break wholeSearchLoop;


                CorePath<V, E> nextCheapestPath;

                // we stop expansion after the walked paths limit and simly get the paths available so far. This is faster
                if(bExpandedPathsOverLimit || getCurrentNumberOfExpandedPaths() >= m_interruptConfig.getWalkedPathsCount4Interrupt())
                {
                    bExpandedPathsOverLimit = true;
                    // simply get what is there
                    nextCheapestPath = m_currentUsedPathExpander.getNextCheapestPath();
                }
                else
                    nextCheapestPath = m_currentUsedPathExpander.getNextCheapestPathAndExpand();



                if(nextCheapestPath != null)
                {
                    m_iWalkedPathCount++;
                    combineWithOtherPathExpandersPaths(nextCheapestPath);
                }
                else
                {
                    // finished
                    itNonFinishedPathExpanders.remove();
                }
            }

        }




        return m_tsSortedResults;
    }



    /**
     * Searches for subgraphs connecting given start nodes (start node set length is 1).
     */
    @SuppressWarnings("unchecked")
    @Override
    public SortedSet<SubGraph<V, E>> search(V... vertices)
    {

        List<Set<V>> startNodeSets = new LinkedList<>();

        for (V vertex : vertices)
            startNodeSets.add(Collections.singleton(vertex));


        return search(startNodeSets);
    }






    public void setResultEqualityWithoutRoots(boolean differentRootIsEqual)
    {
        this.m_bDifferentRootIsEqual = differentRootIsEqual;
    }




}
