package de.dfki.km.jade.graph.recommendation;



import java.util.Collections;
import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Set;
import java.util.SortedSet;
import java.util.TreeSet;

import de.dfki.km.jade.graph.BiGraph;
import de.dfki.km.jade.graph.Edge;
import de.dfki.km.jade.graph.Vertex;
import de.dfki.km.jade.graph.core.CorePath;
import de.dfki.km.jade.graph.core.WeightedEntityContainer;
import de.dfki.km.jade.graph.search.EntitySearch;
import de.dfki.km.jade.graph.search.WalkedPathsInterruptConfig;
import de.dfki.km.jade.graph.search.WeightCalculator;
import de.dfki.km.jade.graph.search.WeightedPathExpander;
import de.dfki.km.jade.util.SaveNonEqualSetObjectsComparator;



/**
 * An entity search implementation, searching for nodes that are in cheap neighbourhood to given start node sets. Nodes that are in cheap neigbourhood
 * to several start node sets will be priorized to these that are only in a neighbourhood to one start node set.
 * 
 * @param <V> the vertex type
 * @param <E> the edge type
 */
public class VertexRecommender<V extends Vertex, E extends Edge<V>> implements EntitySearch<V, E, WeightedEntityContainer<V>>
{
    protected class PathWeight2FinalWeight
    {
        public float fFinalWeight;

        public float fPathWeight;



        public PathWeight2FinalWeight(float fWeight, float fFinalWeight)
        {
            super();
            this.fPathWeight = fWeight;
            this.fFinalWeight = fFinalWeight;
        }
    }


    protected boolean m_bDifferentRootIsEqual = false;

    protected WeightedPathExpander<V, E> m_currentUsedPathExpander;

    protected BiGraph<V, E> m_graph;

    protected WalkedPathsInterruptConfig m_interruptConfig;

    protected int m_iWalkedPathCount = 0;

    protected LinkedList<WeightedPathExpander<V, E>> m_llPathExpanders4StartNodeSets;



    protected HashMap<Vertex, PathWeight2FinalWeight> m_resultVertex2cheapestPathWeight;


    protected TreeSet<WeightedEntityContainer<V>> m_tsSortedResults;



    protected WeightCalculator<V, E, WeightedEntityContainer<V>> m_weightCalculator;



    public VertexRecommender(BiGraph<V, E> graph, WeightCalculator<V, E, WeightedEntityContainer<V>> weightCalculator,
            WalkedPathsInterruptConfig interruptConfig)
    {
        m_graph = graph;

        m_weightCalculator = weightCalculator;

        m_interruptConfig = interruptConfig;
    }




    public int getCurrentNumberOfExpandedPaths()
    {
        int iNumberOfExpandedPaths = 0;

        for (WeightedPathExpander<V, E> pathExpander : m_llPathExpanders4StartNodeSets)
            iNumberOfExpandedPaths += pathExpander.getNumberOfWalkedPaths();


        return iNumberOfExpandedPaths;
    }



    @Override
    public int getCurrentNumberOfWalkedPaths()
    {
        return m_iWalkedPathCount;
    }



    @Override
    public int getCurrentResultCount()
    {
        return m_tsSortedResults.size();
    }







    @Override
    public TreeSet<WeightedEntityContainer<V>> getCurrentResults()
    {
        return m_tsSortedResults;
    }



    @Override
    public BiGraph<V, E> getGraph()
    {
        return m_graph;
    }




    protected int getPathExpandersFoundPathsWithEndCount(V endVertex)
    {
        int iCount = 0;

        for (WeightedPathExpander<V, E> pathExpander : m_llPathExpanders4StartNodeSets)
        {
            if(pathExpander.getCheapestPathsFoundWithEnd(endVertex) != null) iCount++;
        }

        return iCount;
    }



    /**
     * Searches for nodes that are in cheap neighbourhood to given start node sets.
     */
    public SortedSet<WeightedEntityContainer<V>> search(List<Set<V>> startNodeSets)
    {

        m_tsSortedResults = new TreeSet<WeightedEntityContainer<V>>(new SaveNonEqualSetObjectsComparator<WeightedEntityContainer<V>>());
        m_llPathExpanders4StartNodeSets = new LinkedList<>();
        m_resultVertex2cheapestPathWeight = new HashMap<>();


        if(startNodeSets.size() == 0) return m_tsSortedResults;


        for (Set<V> startNodeSet : startNodeSets)
            m_llPathExpanders4StartNodeSets.add(new WeightedPathExpander<>(startNodeSet, m_graph, m_weightCalculator));



        m_interruptConfig.startSearch();


        LinkedList<WeightedPathExpander<V, E>> llNonFinishedPathExpanders = new LinkedList<>(m_llPathExpanders4StartNodeSets);

        boolean bExpandedPathsOverLimit = false;

        wholeSearchLoop: while (llNonFinishedPathExpanders.size() > 0)
        {

            Iterator<WeightedPathExpander<V, E>> itNonFinishedPathExpanders = llNonFinishedPathExpanders.iterator();
            pathExpanderLoop: while (itNonFinishedPathExpanders.hasNext())
            {
                m_currentUsedPathExpander = itNonFinishedPathExpanders.next();


                if(m_interruptConfig.interrupt(this)) break wholeSearchLoop;


                CorePath<V, E> nextCheapestPath;

                // we stop expansion after the walked paths limit and simly get the paths available so far. This is faster
                if(bExpandedPathsOverLimit || getCurrentNumberOfExpandedPaths() >= m_interruptConfig.getWalkedPathsCount4Interrupt())
                {
                    bExpandedPathsOverLimit = true;
                    // simply get what is there
                    nextCheapestPath = m_currentUsedPathExpander.getNextCheapestPath();
                }
                else
                    nextCheapestPath = m_currentUsedPathExpander.getNextCheapestPathAndExpand();



                if(nextCheapestPath == null)
                {
                    // finished
                    itNonFinishedPathExpanders.remove();
                }
                else
                {
                    m_iWalkedPathCount++;
                    V resultVertex = nextCheapestPath.getEndVertex();



                    // Anfrageknoten sind raus
                    for (Set<V> startNodeSet : startNodeSets)
                        if(startNodeSet.contains(resultVertex)) continue pathExpanderLoop;



                    WeightedEntityContainer<V> resultContainer = new WeightedEntityContainer<>(resultVertex);

                    // wir gewichten einen Knoten, welcher mehrfach gefunden wurde, besser (das sind die Sahnestückchen)


                    // wir nehmen immer den billigsten Pfad
                    PathWeight2FinalWeight pathWeight2FinalWeight = m_resultVertex2cheapestPathWeight.get(resultVertex);

                    if(pathWeight2FinalWeight == null)
                    {
                        // der Knoten war noch nie da, neuer Container
                        m_resultVertex2cheapestPathWeight.put(resultVertex,
                                new PathWeight2FinalWeight(nextCheapestPath.getWeight(), nextCheapestPath.getWeight()));

                        resultContainer.setWeight(nextCheapestPath.getWeight());

                        if(!m_weightCalculator.filterResult(resultContainer, m_tsSortedResults)) m_tsSortedResults.add(resultContainer);

                    }
                    else
                    {
                        // war schon mal da
                        float fOldCheapestPathWeight = pathWeight2FinalWeight.fPathWeight;
                        float fOldFinalWeight = pathWeight2FinalWeight.fFinalWeight;

                        pathWeight2FinalWeight.fPathWeight = Math.min(nextCheapestPath.getWeight(), fOldCheapestPathWeight);
                        int iSameEndCount = getPathExpandersFoundPathsWithEndCount(resultVertex);
                        pathWeight2FinalWeight.fFinalWeight = pathWeight2FinalWeight.fPathWeight / iSameEndCount;

                        resultContainer.setWeight(pathWeight2FinalWeight.fFinalWeight);

                        if(!m_weightCalculator.filterResult(resultContainer, m_tsSortedResults))
                        {
                            // der Container wird von remove nur mit dem Gewicht gefunden, mit dem er eingefügt wurde
                            m_tsSortedResults.remove(new WeightedEntityContainer<V>(resultVertex, fOldFinalWeight));
                            m_tsSortedResults.add(resultContainer);
                        }

                    }

                }
            }

        }


        return m_tsSortedResults;
    }



    /**
     * Searches for nodes that are in cheap neighbourhood to given start nodes (start node set length is 1).
     */
    @SuppressWarnings("unchecked")
    @Override
    public SortedSet<WeightedEntityContainer<V>> search(V... vertices)
    {

        List<Set<V>> startNodeSets = new LinkedList<>();

        for (V vertex : vertices)
            startNodeSets.add(Collections.singleton(vertex));


        return search(startNodeSets);
    }







    public void setResultEqualityWithoutRoots(boolean differentRootIsEqual)
    {
        this.m_bDifferentRootIsEqual = differentRootIsEqual;
    }







}
