package de.dfki.km.jade.graph.search;



import java.util.List;
import java.util.Set;
import java.util.SortedSet;

import de.dfki.km.jade.graph.Edge;
import de.dfki.km.jade.graph.Entity;
import de.dfki.km.jade.graph.Graph;
import de.dfki.km.jade.graph.Vertex;


/**
 * The common interface for searching something inside a graph. This could be e.g. connecting paths, subgraphs, or near(recommended) nodes
 * 
 * @author reuschling
 *
 * @param <V> the vertex type
 * @param <E> the edge type
 * @param <R> the result type
 */
public interface EntitySearch<V extends Vertex, E extends Edge<V>, R extends Entity>
{

    public Graph<V, E> getGraph();



    /**
     * Gets the results found so far. This is a snapshot view that is avaliable during the search.
     * 
     * @return the results found so far. This is a snapshot view that is avaliable during the search.
     */
    public Set<R> getCurrentResults();
    
    
    /**
     * Gets the result count found so far. This is a snapshot view that is avaliable during the search.
     * 
     * @return the result count found so far. This is a snapshot view that is avaliable during the search.
     */
    public int getCurrentResultCount();




    /**
     * Performs a search, starting with a list of given vertices.
     * 
     * @param vertices the start nodes/vertices
     * 
     * @return all results, e.g. subgraphs connecting the given vertices
     */
    @SuppressWarnings("unchecked")
    public SortedSet<R> search(V ... vertices);



    /**
     * Performs a search, starting with a list of given node sets.
     * 
     * @param startNodeSets the start node sets
     * 
     * @return all results, e.g. subgraphs connecting the given node sets
     */
    public SortedSet<R> search(List<Set<V>> startNodeSets);



    /**
     * Gets the number of walked paths so far. This is a snapshot view that is avaliable during the search.
     * 
     * @return the number of walked paths so far. This is a snapshot view that is avaliable during the search.
     */
    public int getCurrentNumberOfWalkedPaths();
    
    
    
    
}
