package de.dfki.km.jade.graph.search;



import java.util.HashSet;
import java.util.List;
import java.util.Set;

import de.dfki.km.jade.graph.Edge;
import de.dfki.km.jade.graph.Entity;
import de.dfki.km.jade.graph.Graph;
import de.dfki.km.jade.graph.Vertex;
import de.dfki.km.jade.util.CollectionTool;



/**
 * Weights according the sum of the edge weights of a path. Filters out (sub)graph results where a node appears more often than once. Thus, all
 * paths with 'supporting nodes' will be filtered out. These nodes are not intuitive for the users, depending on your usage scenario. Further, you can
 * filter out similar results according to a fuzzy similarity degree. This is e.g. based on an edit distance - you can filter out all result
 * subgraphs/paths for which an similar one already exists, e.g. one that differs only by one node
 * 
 * @author reuschling
 * 
 * @param <V> the vertex type
 * @param <E> the edge type
 * @param <R> the result type
 */
public class ShrinkingWeightCalculator<V extends Vertex, E extends Edge<V>> extends EdgeSumWeightCalculator<V, E, SubGraph<V, E>>
{

    public static enum SimilarityType {
        /**
         * see {@link CollectionTool#diceCoefficientSimilarity(java.util.Collection, java.util.Collection)} This method works for arbritrary subGraphs
         * results.
         */
        DICE_COEFFICIENT,
        /**
         * see {@link CollectionTool#levenshteinDistance(List, List)}<br>
         * <b>CAUTION :</b> This similarity type works only with subGraphs with one or two internal corepaths, because of regarding the
         * list/'path'positions. In the case you deal with arbritrary subgraph results, and not only with paths, use the {@link #DICE_COEFFICIENT}
         * implementation.
         */
        LEVENSHTEIN_EDIT_DISTANCE,
        /**
         * see {@link CollectionTool#levenshteinSimilarity(List, List)} <br>
         * <b>CAUTION :</b> This similarity type works only with subGraphs with one or two internal corepaths, because of regarding the
         * list/'path'positions. In the case you deal with arbritrary subgraph results, and not only with paths, use the {@link #DICE_COEFFICIENT}
         * implementation.
         */
        LEVENSHTEIN_SIMILARITY
    }


    protected boolean m_bFilterDoubleEntryGraphs = true;

    protected float m_dDistance4filter = -1;




    protected SimilarityType similarityType = SimilarityType.DICE_COEFFICIENT;



    @SuppressWarnings("unchecked")
    @Override
    public boolean filterResult(SubGraph<V, E> potentialResult, Set<SubGraph<V, E>> currentResults)
    {
        // hier wollen wir die Pfade/Subgraphs filtern, in denen ein Knoten mehr als einmal vorkommt. Damit filtern wir die Graphen mit 'Stützknoten'
        // raus, da diese oftmals unintuitiv für einen Nutzer sind

        if(!m_bFilterDoubleEntryGraphs && m_dDistance4filter < 0) return false;


        List<V> lVertices = potentialResult.getVertices();


        if(m_bFilterDoubleEntryGraphs)
        {
            HashSet<V> hsVertices = new HashSet<V>(lVertices);

            if(hsVertices.size() != lVertices.size()) return true;
        }



        if(m_dDistance4filter < 0) return false;


        // wenn es mehr als 2 interne pfade gibt, dann steigen wir auch aus, wenn es keinen Sinn macht
        if(similarityType != SimilarityType.DICE_COEFFICIENT)
        {
            if(((SubGraph<V, E>) potentialResult).getPaths().size() > 2) return false;
        }

        // jetzt gehen wir durch die currentResults, schauen jeweils auf die similarity zwischen einem entry und unserem momentanen Ergebniss, und
        // filtern es raus, sobald es einmal ziemlich ähnlich ist
        for (Entity finalResult : currentResults)
        {
            Graph<V, E> finalResultAsGraph = (Graph<V, E>) finalResult;
            if(similarityType == SimilarityType.DICE_COEFFICIENT)
            {
                double dSimilarity = CollectionTool.diceCoefficientSimilarity(lVertices, finalResultAsGraph.getVertices());
                if((1 - dSimilarity) <= m_dDistance4filter) return true;
            }
            else if(similarityType == SimilarityType.LEVENSHTEIN_SIMILARITY)
            {
                double dSimilarity = CollectionTool.levenshteinSimilarity(lVertices, finalResultAsGraph.getVertices());
                if((1 - dSimilarity) <= m_dDistance4filter) return true;
            }
            else if(similarityType == SimilarityType.LEVENSHTEIN_EDIT_DISTANCE)
            {
                int iDistance = CollectionTool.levenshteinDistance(lVertices, finalResultAsGraph.getVertices());
                if(iDistance <= m_dDistance4filter) return true;
            }
        }



        return false;
    }



    /**
     * Gets the distance under which results will be filtered out.
     * 
     * @return the distance under which results will be filtered out.
     */
    public float getFuzzyFilterDistance()
    {
        return this.m_dDistance4filter;
    }



    /**
     * Gets the used fuzzy filter similarity implementation type
     * 
     * @return the used fuzzy filter similarity implementation type
     */
    public SimilarityType getFuzzyFilterSimilarityType()
    {
        return similarityType;
    }




    public void setFilterDoubleNodeEntryGraphs(boolean bFilterDoubleEntryGraphs)
    {
        this.m_bFilterDoubleEntryGraphs = bFilterDoubleEntryGraphs;

    }



    /**
     * If a result found it will be checked against the already calculated results. In the case it is close to one of them, it will be filtered out.
     * Here you can specify the 'closeness', the distance when possible results will be filtered.
     * 
     * @param distance4filter range: [0...1] <0 means no result will be filtered, which is the default. 0 means only double entries regarding the
     *            vertices positions will be filtered. 1 means all will be filtered out. In the case the similarity type is
     *            {@link SimilarityType#LEVENSHTEIN_EDIT_DISTANCE}, this parameter specifies the allowed edit distance (whole numbers, range [0 ... no
     *            gab]
     * @param similarityType the type of similarity implementation for calculating the distance
     */
    public void setFuzzyFilterDistance(float distance4filter, SimilarityType similarityType)
    {
        this.m_dDistance4filter = distance4filter;
        this.similarityType = similarityType;
    }






}
