package de.dfki.km.jade.graph.search;



import java.util.Collection;
import java.util.List;

import de.dfki.km.jade.graph.Edge;
import de.dfki.km.jade.graph.HierarchicalGraph;
import de.dfki.km.jade.graph.Vertex;
import de.dfki.km.jade.graph.WeightedEntity;
import de.dfki.km.jade.graph.core.CorePath;



/**
 * Represents a subgraph, which means that the internal vertex indices are created from/valid against a 'wrapping' superior graph structure that
 * manages its vertex indices. This interface is e.g. used as a search result. A SubGraph consists of a root vertex with some appended paths or in
 * other words: some paths that share a common end vertex, which is the root node.
 * 
 * @author reuschling
 * 
 * @param <V> the vertex type
 * @param <E> the edge type
 */
public interface SubGraph<V extends Vertex, E extends Edge<V>> extends HierarchicalGraph<V, E>, WeightedEntity
{

    /**
     * Adds all internal paths of a given subGraph
     * 
     * @param subGraph the given subGraph
     */
    public void addGraphPaths(SubGraph<V, E> subGraph);



    /**
     * Adds a path to the {@link SubGraph}. For a legal, consistent state this path should have the root node as end vertex.
     * 
     * @param path the path that should be added to this subGraph
     */
    public void addPath(CorePath<V, E> path);



    /**
     * Adds several paths to this subGraph
     * 
     * @param paths the paths that should be added to this subGraph
     */
    void addPaths(Collection<CorePath<V, E>> paths);



    /**
     * Gets the internal paths this subGraph consists of
     * 
     * @return the internal paths this subGraph consists of
     */
    public List<CorePath<V, E>> getPaths();


}
