package de.dfki.km.jade.graph.search;



import java.util.logging.Logger;

import de.dfki.km.jade.graph.multisteinertree.MultiSteinerTree;
import de.dfki.km.jade.graph.recommendation.VertexRecommender;
import de.dfki.km.jade.util.StringUtils;



/**
 * Interrupts the search according to two criterias:<br>
 * - if a specified number of results are reached => this ensures short-time searches<br>
 * - if a specified number of walked paths is reached. These are normally the paths that will be evaluated as possible result candidates.
 * 
 * @author reuschling
 */
@SuppressWarnings("rawtypes")
public class WalkedPathsInterruptConfig implements SearchInterruptConfig
{

    protected int m_iWalkedPathsCount4Interrupt;

    protected int m_iResultCount4Interrupt;




    /**
     * Constructs a new {@link WalkedPathsInterruptConfig}
     * 
     * @param resultCount4Interrupt reaching more or this number of search results, the search will be interrupted
     * @param walkedPathsCount4Interrupt reaching more or this number of walked paths during graph traversation, the search will be interrupted,
     *            independently how many results were found at this time
     */
    public WalkedPathsInterruptConfig(int resultCount4Interrupt, int walkedPathsCount4Interrupt)
    {

        m_iResultCount4Interrupt = resultCount4Interrupt;

        m_iWalkedPathsCount4Interrupt = walkedPathsCount4Interrupt;

    }



    /**
     * Reaching more or this number of walked paths during graph traversation, the search will be interrupted, independently how many results were
     * found at this time
     * 
     * @return reaching more or this number of walked paths during graph traversation, the search will be interrupted, independently how many results
     *         were found at this time
     */
    public int getWalkedPathsCount4Interrupt()
    {
        return m_iWalkedPathsCount4Interrupt;
    }




    /**
     * Reaching more or this number of search results, the search will be interrupted
     * 
     * @return reaching more or this number of search results, the search will be interrupted
     */
    public int getResultCount4Interrupt()
    {
        return m_iResultCount4Interrupt;
    }




    @Override
    public boolean interrupt(EntitySearch entitySearch)
    {




        // wenn wir die Anzahl der gewünschten Ergebnisse haben, brechen wir glücklich ab.
        int iCurrentResultCount = entitySearch.getCurrentResultCount();
        if(iCurrentResultCount >= m_iResultCount4Interrupt)
        {

            int iWalkedPathCount = entitySearch.getCurrentNumberOfWalkedPaths();
            StringBuilder strbMessage = new StringBuilder();
            strbMessage.append("Configured result count of ").append(StringUtils.beautifyNumber(m_iResultCount4Interrupt)).append(" reached (")
                    .append(StringUtils.beautifyNumber(iCurrentResultCount)).append("). Search will be interrupted. Walked ")
                    .append(StringUtils.beautifyNumber(iWalkedPathCount)).append(" paths.");

            if(entitySearch instanceof VertexRecommender)
                strbMessage.append(" Expanded ").append(StringUtils.beautifyNumber(((VertexRecommender) entitySearch).getCurrentNumberOfExpandedPaths()))
                        .append(" paths.");
            if(entitySearch instanceof MultiSteinerTree)
                strbMessage.append(" Expanded ").append(StringUtils.beautifyNumber(((MultiSteinerTree) entitySearch).getCurrentNumberOfExpandedPaths()))
                        .append(" paths.");


            Logger.getLogger(WalkedPathsInterruptConfig.class.getName()).info(strbMessage.toString());

            return true;
        }


        // wenn wir eine gewisse Anzahl an Pfaden durchlaufen haben, dann ist das das definierte Ende. Dieser lookup zappt jetzt auch :)
        int iWalkedPathCount = entitySearch.getCurrentNumberOfWalkedPaths();
        if(iWalkedPathCount >= m_iWalkedPathsCount4Interrupt)
        {

            StringBuilder strbMessage = new StringBuilder();
            strbMessage.append("Configured path count to walk of ").append(StringUtils.beautifyNumber(m_iWalkedPathsCount4Interrupt)).append(" reached (")
                    .append(StringUtils.beautifyNumber(iWalkedPathCount)).append("). Search will be interrupted. Found ")
                    .append(StringUtils.beautifyNumber(iCurrentResultCount)).append(" results.");

            if(entitySearch instanceof VertexRecommender)
                strbMessage.append(" Expanded ").append(StringUtils.beautifyNumber(((VertexRecommender) entitySearch).getCurrentNumberOfExpandedPaths()))
                        .append(" paths.");
            if(entitySearch instanceof MultiSteinerTree)
                strbMessage.append(" Expanded ").append(StringUtils.beautifyNumber(((MultiSteinerTree) entitySearch).getCurrentNumberOfExpandedPaths()))
                        .append(" paths.");


            Logger.getLogger(WalkedPathsInterruptConfig.class.getName()).info(strbMessage.toString());

            return true;
        }




        return false;
    }





    /**
     * Reaching more or this number of walked paths during graph traversation, the search will be interrupted, independently how many results were
     * found at this time
     * 
     * @param constructedPathsCount4Interrupt reaching more or this number of walked paths during graph traversation, the search will be interrupted,
     *            independently how many results were found at this time
     */
    public void setWalkedPathsCount4Interrupt(int constructedPathsCount4Interrupt)
    {
        this.m_iWalkedPathsCount4Interrupt = constructedPathsCount4Interrupt;
    }




    /**
     * Reaching more or this number of search results, the search will be interrupted
     * 
     * @param resultCount4Interrupt reaching more or this number of search results, the search will be interrupted
     */
    public void setResultCount4Interrupt(int resultCount4Interrupt)
    {
        m_iResultCount4Interrupt = resultCount4Interrupt;

    }



    @Override
    public void startSearch()
    {
        // NOP
    }



}
