package de.dfki.km.jade.graph.search;



import java.util.Set;

import de.dfki.km.jade.graph.Edge;
import de.dfki.km.jade.graph.Entity;
import de.dfki.km.jade.graph.Graph;
import de.dfki.km.jade.graph.Path;
import de.dfki.km.jade.graph.Vertex;
import de.dfki.km.jade.graph.core.CorePath;
import de.dfki.km.jade.graph.multisteinertree.MultiSteinerTree;



/**
 * A {@link WeightCalculator} evaluates a path that was founded during graph traveral by calculating an according weight. This weight will be further
 * used for e.g. making the choice which paths/nodes should be traversed in the next traversal step, or also to score path-based results. Further, as
 * a not so time critical possibility to take control, a {@link WeightCalculator} also filters out potential search results that should be not
 * considered as final. This is a last step choice.<br>
 * The {@link WeightCalculator} represents the heard of all searches, and gives the algorithms their characteristics. You can e.g. implement a
 * shortest path algorithm by simply wheighting the paths according to their length. You can implement someting like an A* algorithm by specifying
 * further criterias, if the search implementation considers the calculated weight during traversal (as {@link MultiSteinerTree} does. You can makey
 * hybrids. etc.<br>
 * With the filter function, you are e.g. able to only consider subGraph results with path characteristic, you can filter out double entries or
 * similar results according to some criteria, as {@link ShrinkingWeightCalculator} does.<br>
 * Simply make your own mixture. We hope you enjoy.
 * 
 * @author reuschling
 * 
 * @param <V> the vertex type
 * @param <E> the edge type
 * @param <R> the result type
 */
public abstract class WeightCalculator<V extends Vertex, E extends Edge<V>, R extends Entity>
{
    /**
     * Calculates the weight and sets it inside the given path. In the case the path should not considered further for searching at all (traversal
     * cutting), the method returns false and set the weight to {@link Float#MAX_VALUE}.<br>
     * This method will be invoked during graph traversion and is highly performance critical! In the case it is enough to filter potential results,
     * think about using {@link #filterResult(Graph)}.
     * 
     * 
     * @param corePath the path we want to have a weight for. Note that the default implementation of
     *            {@link WeightCalculator#calculateAndSetWeight(CorePath)} uses
     *            {@link WeightCalculator#calculateWeight(Vertex, Edge, Boolean, CorePath)}, thus assuming that the weight of the parent corePath
     *            inside corePath is already set.
     * 
     * @return true in the case the path should be considered further for searching, false otherwise (traversal cutting)
     */
    public boolean calculateAndSetWeight(CorePath<V, E> corePath)
    {
        float weight = calculateWeight(corePath.getEndVertex(), corePath.getEndEdge(), corePath.getParent());

        corePath.setWeight(weight);

        if(weight == Float.MAX_VALUE) return false;


        return true;
    }



    /**
     * Calculates the weight of a potential new path. In the case this path should not considered further for searching at all (traversal cutting),
     * the method returns {@link Float#MAX_VALUE}. The potential new path consists of an existing parent path plus an edge with vertex as postfix.
     * This method can be used to check a potential path before finally creating it. In the case you have a path Object yet, you can use the
     * {@link #calculateAndSetWeight(CorePath)} method.<br>
     * This method will be invoked during graph traversion and is highly performance critical! In the case it is enough to filter potential results,
     * think about using {@link #filterResult(Graph)}.
     * 
     * @param newEndVertex the new, postfix end Vertex
     * @param newEndEdge the new, postfix end Edge
     * @param parentPath the prefix/parent path. This one already HAS a set weight. Use it as input for calculating the weight of the resulting, new
     *            path (parentPath + newEndVertex/newEndEdge.
     * 
     * @return the calculated weight, {@link Float#MAX_VALUE} in the case the path should not considered further during the search (traversal cutting)
     */
    abstract public float calculateWeight(V newEndVertex, E newEndEdge, CorePath<V, E> parentPath);



    /**
     * After traversing, when a result is found, this method is executed to check whether this result should be considered as a final result or not.
     * This method will be not performed during graph traversal, thus it is NOT such performance critical as
     * {@link #calculateWeight(Vertex, Edge, Boolean, CorePath)} and {@link #calculateAndSetWeight(CorePath)}.
     * 
     * @param potentialResult a founded result that should be checked for filtering. Can e.g. be a {@link SubGraph}, {@link Path}, {@link Vertex} or
     *            {@link CorePath}, depending on the search implementation.
     * @param currentResults the results found so far. Can be used as criteria for filtering the current result, maybe for filtering out redundant
     *            'double' entries.
     * 
     * @return true in the case this result should NOT considered as final result. False otherwise.
     */
    abstract public boolean filterResult(R potentialResult, Set<R> currentResults);







}
