package de.dfki.km.jade.graph.search;



import java.util.Collection;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.PriorityQueue;

import de.dfki.km.jade.graph.BiGraph;
import de.dfki.km.jade.graph.Edge;
import de.dfki.km.jade.graph.Entity;
import de.dfki.km.jade.graph.Vertex;
import de.dfki.km.jade.graph.core.CorePath;



/**
 * This class simply expands given start nodes, creating paths by concatenating neighbour nodes, stepwise. Each resulting path will be weighted by a
 * given {@link WeightCalculator}. Cheap paths with small weight will be expanded and returned at each traversal step first.
 * 
 * 
 * @param <V> the vertex type
 * @param <E> the edge type
 */
public class WeightedPathExpander<V extends Vertex, E extends Edge<V>>
{

    protected BiGraph<V, E> m_graph;

    protected HashMap<V, List<CorePath<V, E>>> m_hsPathEndVertex2AccordingCheapestPaths = new HashMap<>();

    // better performance could be gained with e.g. a Van Emde Boas based (depends,tested now)or with a Soft Heap based PriorityQueue implementation
    protected PriorityQueue<CorePath<V, E>> m_pqSortedExpandedPaths = new PriorityQueue<>();

    protected int m_iCheapestPathFoundCount = 0;

    protected WeightCalculator<V, E, ? extends Entity> m_weightCalculator;






    public WeightedPathExpander(Collection<V> startNodes, BiGraph<V, E> graph, WeightCalculator<V, E, ? extends Entity> weightCalculator)
    {

        m_graph = graph;

        m_weightCalculator = weightCalculator;

        for (V node : startNodes)
            m_pqSortedExpandedPaths.add(new CorePath<V, E>(node));

    }





    /**
     * Gets all paths found so far with {@link #getNextCheapestPathAndExpand()} invocations that ends with a given vertex
     * 
     * @param endVertex the end vertex
     * 
     * @return all paths found so far with {@link #getNextCheapestPathAndExpand()} invocations that ends with a given vertex
     */
    public List<CorePath<V, E>> getCheapestPathsFoundWithEnd(V endVertex)
    {
        return m_hsPathEndVertex2AccordingCheapestPaths.get(endVertex);
    }



    /**
     * Gets the current cheapest (smallest weight) path, without expanding further neighbour paths. Returns null in the case no path is left in the
     * current queue. Once found and returned, the path is available for lookup with {@link #getCheapestPathsFoundWithEnd(Vertex)}
     * 
     * @return the next, cheapest path (with smallest weight). Null in the case there are no paths left.
     */
    public CorePath<V, E> getNextCheapestPath()
    {

        // remove current Vertex from expanded paths list
        CorePath<V, E> cheapestPath = m_pqSortedExpandedPaths.poll();

        if(cheapestPath == null) return null;


        V cheapestPathEndVertex = cheapestPath.getEndVertex();


        // save cheapest path for lookup
        List<CorePath<V, E>> lPathsWithSameEndVertexFound = m_hsPathEndVertex2AccordingCheapestPaths.get(cheapestPathEndVertex);

        if(lPathsWithSameEndVertexFound == null)
        {
            lPathsWithSameEndVertexFound = new LinkedList<>();
            m_hsPathEndVertex2AccordingCheapestPaths.put(cheapestPathEndVertex, lPathsWithSameEndVertexFound);
        }

        lPathsWithSameEndVertexFound.add(cheapestPath);
        m_iCheapestPathFoundCount++;



        return cheapestPath;
    }





    /**
     * Gets the current cheapest (smallest weight) path, expanding it with all neighbour nodes before returning it. The resulting paths will be
     * weighted and maybe filtered. Returns null in the case no path is left - the expansion is finished, there is no path left to return (graph is
     * completely traversed or all neighbour paths are filtered out due to some (user defined) criteria. Once found and returned, the path is
     * available for lookup with {@link #getCheapestPathsFoundWithEnd(Vertex)}
     * 
     * @return the next, cheapest path (with smallest weight). Null in the case there are no paths left.
     */
    public CorePath<V, E> getNextCheapestPathAndExpand()
    {

        CorePath<V, E> cheapestPath = getNextCheapestPath();

        if(cheapestPath == null) return null;


        V cheapestPathEndVertex = cheapestPath.getEndVertex();


        // expand with all neighbours if not filtered. Weighted path will be stored in sorted paths member.
        for (E edge : m_graph.getEdgesWithTarget(cheapestPathEndVertex))
            maybeCreateNewPath(edge.getSource(), edge, cheapestPath);

        for (E edge : m_graph.getEdgesWithSource(cheapestPathEndVertex))
            maybeCreateNewPath(edge.getTarget(), edge, cheapestPath);



        return cheapestPath;
    }



    /**
     * Gets the current number of cheapest paths found by {@link #getNextCheapestPathAndExpand()} invocations.
     * 
     * @return the current number of cheapest paths found by {@link #getNextCheapestPathAndExpand()} invocations.
     */
    public int getNumberOfCheapestPathsFound()
    {
        return m_iCheapestPathFoundCount;
    }







    /**
     * This is the number of cheapest paths returned/found, plus the current expanded path count, both calculated by past
     * {@link #getNextCheapestPathAndExpand()} invocations
     * 
     * @return the number of currently walked paths
     */
    public int getNumberOfWalkedPaths()
    {
        return m_pqSortedExpandedPaths.size() + m_iCheapestPathFoundCount;
    }



    protected void maybeCreateNewPath(V newEndVertex, E newEndEdge, CorePath<V, E> parentPath)
    {
        // keine zykel
        if(parentPath.contains(newEndVertex)) return;


        float fWeight = m_weightCalculator.calculateWeight(newEndVertex, newEndEdge, parentPath);

        // Pfad wird für weitere Verarbeitung ausgeschlossen
        if(fWeight == Float.MAX_VALUE) return;


        // wir machen einen neuen Pfad
        CorePath<V, E> newPath = new CorePath<>(newEndVertex, newEndEdge, parentPath);

        newPath.setWeight(fWeight);


        m_pqSortedExpandedPaths.add(newPath);
    }


}
