package de.dfki.km.jade.util;



import java.util.Collection;
import java.util.LinkedList;
import java.util.List;

import org.apache.commons.lang3.time.StopWatch;


/**
 * Some utility methods regarding Collections
 * 
 * @author reuschling
 */
public final class CollectionTool
{
    /**
     * Returns the dice coefficient, some degree of 'similarity' between two Collections. It is based on the percentage amount of intersection. Thus,
     * a difference of one entry is not so relevant in longer lists, compared to smaller lists. This differs from the raw Levenshtein edit distance.
     * The main difference against the Levenshtein similarity is, that Levenshtein regards the position of the elements. The dice coefficient only
     * looks on the intersections. The performance is influenced by two things:<br>
     * <li>how fast can col1 be iterated?</li> <li>how fast is the contains(obj) method of col2</li>
     * 
     * @param col1 collection1, should be efficient in iterating its elements
     * @param col2 collection2, should be efficient in its contains(obj) method
     * 
     * @return the resulting dice coefficient. range [0..1], 1 means equality.
     */
    @SuppressWarnings("rawtypes")
    public final static double diceCoefficientSimilarity(Collection col1, Collection col2)
    {
        if(col1 == null)
        {
            if(col2 == null) return 1;
            if(col2.size() == 0) return 1;
        }
        else
        {
            if(col2 == null && col1.size() == 0) return 1;
        }

        int iIntersectionCount = 0;
        for (Object obj1 : col1)
            if(col2.contains(obj1)) iIntersectionCount++;


        return (2 * iIntersectionCount) / (col1.size() + col2.size());
    }



    /**
     * Gives the Levenshtein edit distance between the two given lists. Best performance with fast random access lists.
     * 
     * @param list1 list1
     * @param list2 list2
     * 
     * @return the Levenshtein edit distance between these two lists
     */

    @SuppressWarnings("rawtypes")
    public static int levenshteinDistance(List list1, List list2)
    {
        if(list1 == null || list1.size() == 0)
        {
            if(list2 == null) return 0;

            return list2.size();
        }
        else
        {
            if(list2 == null || list2.size() == 0) return list1.size();
        }



        int s1_len = list1.size();
        int s2_len = list2.size();

        int i1 = 0;
        int i2 = 0;
        int a, b = 0, c;
        Object c2;
        int[] row = new int[s1_len];


        while (i1 < s1_len)
            row[i1] = ++i1;


        while (i2 < s2_len)
        {
            c2 = list2.get(i2);
            a = i2;
            ++i2;
            b = i2;

            for (i1 = 0; i1 < s1_len; ++i1)
            {
                c = a + (list1.get(i1).equals(c2) ? 0 : 1);
                a = row[i1];
                b = b < a ? (b < c ? b + 1 : c) : (a < c ? a + 1 : c);
                row[i1] = b;
            }
        }
        return b;

    }







    /**
     * Gets a Levenshtein-based similarity degree between two given lists. The Levenshtein distance of the two lists will be normalized with the max
     * list size. Thus, as it is the case at the dice coefficient, an edit distance of 1 is less relevant at longer lists. If you don't want this, use
     * the raw edit distance. The main difference against the dice coefficient similarity is, that Levenshtein regards the position of the elements.
     * The dice coefficient only looks on the intersections.
     * 
     * @param list1
     * @param list2
     * 
     * @return a Levenshtein-based similarity degree between two given lists, range [0..1], 1 means equality
     */
    @SuppressWarnings("rawtypes")
    static public double levenshteinSimilarity(List list1, List list2)
    {
        double distance = levenshteinDistance(list1, list2);

        int iMaxListSize = Math.max(list1.size(), list2.size());

        if(iMaxListSize == 0) return 1;


        return 1 - distance / iMaxListSize;
    }



    public static void main(String[] args)
    {
        java.util.ArrayList<String> al1 = null;
        java.util.ArrayList<String> al2 = null;

        System.out.println(levenshteinDistance(al1, al2));
        System.out.println("0 ////////////////");

        al1 = new java.util.ArrayList<>();

        System.out.println(levenshteinDistance(al1, al2));
        System.out.println("0 ////////////////");

        al2 = new java.util.ArrayList<>();

        al1.add("bla");

        System.out.println(levenshteinDistance(al1, al2));
        System.out.println("1 ////////////////");

        al2.add("bla");

        System.out.println(levenshteinDistance(al1, al2));
        System.out.println("0////////////////");

        al1.add("blubb");
        al1.add("blubb1");
        al1.add("blubb2");

        System.out.println(levenshteinDistance(al1, al2));
        System.out.println("3////////////////");

        al2.add("blubb");

        System.out.println(levenshteinDistance(al1, al2));
        System.out.println("2////////////////");

        al2.add("blubb1");

        System.out.println(levenshteinDistance(al1, al2));
        System.out.println("1////////////////");

        al2.add("blubb2");

        System.out.println(levenshteinDistance(al1, al2));
        System.out.println("0////////////////");


        al2.add("blubb23");
        al2.add("blubb23");
        al2.add("blubb23");
        al2.add("blubb23");


        System.out.println(levenshteinDistance(al1, al2));
        System.out.println("4////////////////");

        StopWatch stopWatch = new StopWatch();
        stopWatch.start();

        for (int i = 0; i < 10000000; i++)
        {
            levenshteinDistance(al1, al2);
        }
        stopWatch.stop();
        System.out.println(stopWatch.getTime());


        stopWatch = new StopWatch();
        stopWatch.start();

        for (int i = 0; i < 10000000; i++)
        {
            levenshteinDistance(al1, al2);
        }
        stopWatch.stop();
        System.out.println(stopWatch.getTime());
    }



    public final static LinkedList<Integer> toLinkedIntegerList(final int[] ints)
    {
        final LinkedList<Integer> toReturn = new LinkedList<Integer>();

        for (int i : ints)
        {
            toReturn.add(new Integer(i));
        }

        return toReturn;
    }






}
