package de.dfki.km.jade.util;



import java.util.AbstractList;
import java.util.Arrays;
import java.util.Collection;
import java.util.List;
import java.util.RandomAccess;

import org.apache.commons.lang3.ArrayUtils;



/**
 * An ArrayList implementation that gives the maximum performance for read operations in favour of write operation performance, beside a possible cast
 * overhead in the get method because of the use of generics. The jdk ArrayList maintains an internal size variable, whereas the internal array can be
 * even bigger in reality. This is to avoid too much arraycopy operations for growing the internal array for {@link #add(Object)} invocations.<br>
 * {@link FastReadArrayList} grows everytime to the amount of one during an {@link #add(Object)}, which means that the internal array length
 * corresponds to the real ArrayList size. Thus, a size/length check at the {@link #get(int)} method is not necessary. To get good write performance
 * with this class, {@link #grow(int)} the array length first and then use the {@link #setQuick(int, Object)}. Or use one of the
 * {@link #addAll(Collection)} methods.
 * 
 * @author reuschling
 * @param <E> the edge type
 */
public class FastReadArrayList<E> extends AbstractList<E> implements List<E>, RandomAccess
{

    protected Object[] m_dataArray;




    /**
     * Constructs an empty FastReadSlowWriteArrayList
     */
    public FastReadArrayList()
    {
        m_dataArray = new Object[0];
    }



    /**
     * Constructs a {@link FastReadArrayList} with given content
     * 
     * @param initialContent content for the internal array
     */
    public FastReadArrayList(Collection<? extends E> initialContent)
    {
        m_dataArray = initialContent.toArray();

        if(m_dataArray.getClass() != Object[].class) m_dataArray = Arrays.copyOf(m_dataArray, m_dataArray.length, Object[].class);
    }



    /**
     * Wraps a given array. Not that the reference of the internal array may be changed as a result of array modifications (add or remove operations)
     * 
     * @param array2wrap the array to wrap
     */
    public FastReadArrayList(E[] array2wrap)
    {
        m_dataArray = array2wrap;
    }



    /**
     * Constructs an {@link FastReadArrayList} with an initial size
     * 
     * @param iInitialSize the initial size. The internal array will have null entries
     */
    public FastReadArrayList(int iInitialSize)
    {
        m_dataArray = new Object[iInitialSize];
    }



    /**
     * This is a costly operation, because the array will be grown up by one. In the case you want to add more than one element, grow your array by
     * the number of elements you want to add and then set them with the according indices afterwards, to avoid unnecessary arrayCopies
     * 
     * @param element the element you want to add. Note that this is a very costly operation.
     */
    @Override
    public boolean add(E element)
    {
        m_dataArray = ArrayUtils.add(m_dataArray, element);

        return true;
    }



    /**
     * This is a costly operation, because the array will be grown up by one. In the case you want to add more than one element, grow your array by
     * the number of elements you want to add and then set them with the according indices afterwards, to avoid unnecessary arrayCopies
     * 
     * @param index the oindex where you want to add the new element
     * @param element the element you want to add. Note that this is a very costly operation.
     */
    @Override
    public void add(int index, E element)
    {
        m_dataArray = ArrayUtils.add(m_dataArray, index, element);
    }



    /**
     * Adds new elements to the end of the internal array. This method will perform an arraycopy
     * 
     * @param elements the elements to add
     * 
     * @return true in the case the array was modified
     */
    @Override
    public boolean addAll(Collection<? extends E> elements)
    {
        if(elements.size() == 0) return false;

        m_dataArray = ArrayUtils.addAll(m_dataArray, elements.toArray());

        return true;
    }



    /**
     * Adds new elements to the end of the internal array. This method will perform an arraycopy
     * 
     * @param elements the elements to add
     * 
     * @return true in the case the array was modified
     */
    @SuppressWarnings("unchecked")
    public boolean addAll(E... elements)
    {
        if(elements.length == 0) return false;

        m_dataArray = ArrayUtils.addAll(m_dataArray, elements);

        return true;
    }



    /**
     * Adds new elements to a specified index into the internal array. This method will perform an arraycopy
     * 
     * @param index the index where you want to add the new elements
     * @param elements the elements to add
     * 
     * @return true in the case the array was modified
     */
    @Override
    public boolean addAll(int index, Collection<? extends E> elements)
    {
        if(elements.size() == 0) return false;

        int iMovedElementsCount = this.size() - index;

        if(iMovedElementsCount <= 0)
        {
            return this.addAll(elements);
        }
        else
        {
            Object[] temp = new Object[this.size() + elements.size()];

            System.arraycopy(m_dataArray, 0, temp, 0, index);
            System.arraycopy(m_dataArray, index, temp, index + elements.size(), iMovedElementsCount);

            System.arraycopy(elements, 0, m_dataArray, index, elements.size());

            return true;
        }
    }



    /**
     * Adds new elements to a specified index into the internal array. This method will perform an arraycopy
     * 
     * @param index the index where you want to add the new elements
     * @param elements the elements to add
     * 
     * @return true in the case the array was modified
     */
    @SuppressWarnings("unchecked")
    public boolean addAll(int index, E... elements)
    {
        if(elements.length == 0) return false;

        int iMovedElementsCount = this.size() - index;

        if(iMovedElementsCount <= 0)
        {
            return this.addAll(elements);
        }
        else
        {
            Object[] temp = new Object[this.size() + elements.length];

            System.arraycopy(m_dataArray, 0, temp, 0, index);
            System.arraycopy(m_dataArray, index, temp, index + elements.length, iMovedElementsCount);

            System.arraycopy(elements, 0, m_dataArray, index, elements.length);

            return true;
        }
    }



    @Override
    public void clear()
    {
        m_dataArray = new Object[0];
    }



    @Override
    public boolean contains(Object objectToFind)
    {
        return ArrayUtils.contains(m_dataArray, objectToFind);
    }



    @SuppressWarnings("unchecked")
    @Override
    public E get(int index)
    {
        return (E) m_dataArray[index];
    }



    /**
     * Grows the internal array, if necessary, by adding amount of null values
     * 
     * @param iNewSize the new size of the array
     * 
     * @return true in the case the array lenght was modified
     */
    public boolean grow(int iNewSize)
    {
        int iOldSize = m_dataArray.length;

        if(iOldSize >= iNewSize) return false;

        m_dataArray = ArrayUtils.addAll(m_dataArray, new Object[iNewSize - iOldSize]);

        return true;
    }



    @Override
    public int indexOf(Object objectToFind)
    {
        return ArrayUtils.indexOf(m_dataArray, objectToFind);
    }



    @Override
    public int lastIndexOf(Object valueToFind)
    {
        return ArrayUtils.lastIndexOf(m_dataArray, valueToFind);
    }



    @Override
    public E remove(int index)
    {
        E oldElement = get(index);

        m_dataArray = ArrayUtils.remove(m_dataArray, index);

        return oldElement;
    }



    @Override
    public boolean remove(Object object2remove)
    {
        int iOldSize = m_dataArray.length;
        m_dataArray = ArrayUtils.removeElement(m_dataArray, object2remove);

        if(m_dataArray.length != iOldSize) return true;

        return false;
    }



    @Override
    public boolean removeAll(Collection<?> elements2remove)
    {
        int iOldSize = m_dataArray.length;

        m_dataArray = ArrayUtils.removeElement(m_dataArray, elements2remove);

        return (iOldSize != m_dataArray.length);
    }



    @Override
    public boolean retainAll(Collection<?> c)
    {
        final Object[] elementData = m_dataArray;
        int r = 0;
        int w = 0;
        int size = size();

        boolean modified = false;
        try
        {
            for (; r < size; r++)
                if(c.contains(elementData[r]) == true) elementData[w++] = elementData[r];
        }
        finally
        {
            // Preserve behavioral compatibility with AbstractCollection,
            // even if c.contains() throws.
            if(r != size)
            {
                System.arraycopy(elementData, r, elementData, w, size - r);
                w += size - r;
            }
            if(w != size)
            {
                for (int i = w; i < size; i++)
                    elementData[i] = null;

                modCount += size - w;
                size = w;
                modified = true;
            }
        }


        return modified;
    }



    @Override
    public E set(int index, E element)
    {
        E oldElement = get(index);

        m_dataArray[index] = element;

        return oldElement;
    }



    /**
     * Sets an element to the specified position inside the internal array. Grows the array if necessary.
     * 
     * @param index the index where to store the new element
     * @param element the element
     * 
     * @return the former element
     */
    public E setAndGrow(int index, E element)
    {
        E oldValue;

        if(!grow(index + 1))
            oldValue = get(index);
        else
            oldValue = null;

        setQuick(index, element);


        return oldValue;
    }



    /**
     * Sets an element to the specified position inside the internal array, without returning the former value. Grows the array if necessary.
     * 
     * @param index the index where to store the new element
     * @param element the element
     */
    public void setAndGrowQuick(int index, E element)
    {
        grow(index + 1);
        setQuick(index, element);
    }



    /**
     * Sets an element to the specified position inside the internal array, without returning the former value.
     * 
     * @param index the index where to set
     * @param element the element to set
     */
    public void setQuick(int index, E element)
    {
        m_dataArray[index] = element;
    }



    @Override
    public int size()
    {
        return m_dataArray.length;
    }




    /**
     * Returns just the internal array reference
     */
    @Override
    public Object[] toArray()
    {
        return m_dataArray;
    }




    @SuppressWarnings("unchecked")
    @Override
    public <T> T[] toArray(T[] a)
    {
        if(a.length < size()) return (T[]) Arrays.copyOf(m_dataArray, size(), a.getClass());

        System.arraycopy(m_dataArray, 0, a, 0, size());

        if(a.length > size()) a[size()] = null;

        return a;
    }





}
