package de.dfki.km.jade.util;



import java.text.NumberFormat;
import java.text.ParseException;
import java.util.Locale;
import java.util.regex.MatchResult;
import java.util.regex.Matcher;
import java.util.regex.Pattern;



public class StringUtils
{


    public static void main(String[] args)
    {
        System.out.println(StringUtils.beautifyNumber(1230024.5684f));
    }



    /**
     * Formats numbers according the current default locale. Uses NumberFormat.getInstance(Locale.getDefault()).
     * 
     * @param no2Beautify the number to beautify
     * 
     * @return the beauty
     */
    static public String beautifyNumber(Number no2Beautify)
    {
        if(no2Beautify instanceof Double || no2Beautify instanceof Long)
            return NumberFormat.getInstance(Locale.getDefault()).format(no2Beautify);
        else if(no2Beautify instanceof Float)
            return NumberFormat.getInstance(Locale.getDefault()).format(((Float) no2Beautify).doubleValue());
        else if(no2Beautify instanceof Integer)
            return NumberFormat.getInstance(Locale.getDefault()).format(((Integer) no2Beautify).longValue());
        else
            return beautifyNumber(no2Beautify.toString());
    }



    /**
     * Formats numbers according the current default locale. Uses NumberFormat.getInstance(Locale.getDefault()).
     * 
     * @param strNo2Beautify the number to beautify as String
     * 
     * @return the beauty
     */
    static public String beautifyNumber(String strNo2Beautify)
    {

        NumberFormat numberFormat = NumberFormat.getInstance(Locale.getDefault());
        Number asNumber;
        try
        {
            asNumber = numberFormat.parse(strNo2Beautify);
        }
        catch (ParseException e)
        {
            throw new RuntimeException(e);
        }


        return numberFormat.format(asNumber);
    }



    /**
     * Checks whether a String is not null and - in the case it isn't - has more content than whitespace. This is usefull in the case you wnt to
     * ensure there is at least some (non-whitespace) content inside a String.
     * 
     * @param string2check the String you wants to check
     * 
     * @return true in the case there is some none-whitespace content, false otherwise
     */
    static public boolean nullOrWhitespace(String string2check)
    {
        return !(string2check != null && string2check.trim().length() > 0);
    }



    /**
     * Finds the occurences of a regular expression in a given String
     * 
     * @param strRegEx the regular expression to search
     * @param strInput the input String
     * 
     * @return the Matcher Object, the results can be 'iterated' with 'Matcher.find()' and 'Matcher.toMatchResult();'
     */
    public static Matcher find(String strRegEx, String strInput)
    {
        return Pattern.compile(strRegEx).matcher(strInput);
    }



    /**
     * Gets the first occurence of a regular expression in a given String
     * 
     * @param strRegEx the regular expression to search
     * @param strInput the input String
     * 
     * @return the first occurence, of null in the case there was no match at all
     */
    public static MatchResult findFirst(String strRegEx, String strInput)
    {
        Matcher matcher = Pattern.compile(strRegEx).matcher(strInput);

        if(matcher.find())
            return matcher.toMatchResult();
        else
            return null;
    }



    /**
     * Gets the last occurence of a regular expression in a given String
     * 
     * @param strRegEx the regular expression to search
     * @param strInput the input String
     * 
     * @return the last occurence, of null in the case there was no match at all
     */
    public static MatchResult findLast(String strRegEx, String strInput)
    {
        Matcher matcher = Pattern.compile(strRegEx).matcher(strInput);

        MatchResult lastMatch = null;

        while (matcher.find())
            lastMatch = matcher.toMatchResult();

        return lastMatch;
    }



    /**
     * Checks whether a given String can be parsed as Double
     * 
     * @param string2Check the String you want to check
     * 
     * @return true in the case the String can be parsed as Double
     */
    static public boolean isDouble(String string2Check)
    {
        try
        {
            Double.valueOf(string2Check);

            return true;
        }
        catch (NumberFormatException e)
        {
            return false;
        }
    }



    /**
     * Checks whether a given String can be parsed as Float
     * 
     * @param string2Check the String you want to check
     * 
     * @return true in the case the String can be parsed as Float
     */
    static public boolean isFloat(String string2Check)
    {
        try
        {
            Float.valueOf(string2Check);

            return true;
        }
        catch (NumberFormatException e)
        {
            return false;
        }
    }



    /**
     * Checks whether a given String can be parsed as Integer
     * 
     * @param string2Check the String you want to check
     * 
     * @return true in the case the String can be parsed as Integer
     */
    static public boolean isInteger(String string2Check)
    {
        try
        {
            Integer.valueOf(string2Check);

            return true;
        }
        catch (NumberFormatException e)
        {
            return false;
        }
    }



    /**
     * Checks whether a given String can be parsed as Long
     * 
     * @param string2Check the String you want to check
     * 
     * @return true in the case the String can be parsed as Long
     */
    static public boolean isLong(String string2Check)
    {
        try
        {
            Long.valueOf(string2Check);

            return true;
        }
        catch (NumberFormatException e)
        {
            return false;
        }
    }



    /**
     * Escapes all occurences of the given chars by adding a preceding '\'
     * 
     * @param strOriginString the String you want to escape
     * @param chars2escape the characters you want to escape
     * 
     * @return the escaped String
     */
    static public String escapeChars(String strOriginString, char... chars2escape)
    {
        String strEscapedString = strOriginString;
        for (char char2escape : chars2escape)
            strEscapedString = strEscapedString.replace(String.valueOf(char2escape).toString(), "\\" + char2escape);

        return strEscapedString;
    }



    /**
     * Unescapes all occurences of the given chars by deleting a preceding '\'
     * 
     * @param strEscapedString the String you want to unescape
     * @param chars2unescape the characters you want to unescape
     * 
     * @return the unescaped String
     */
    static public String unescapeChars(String strEscapedString, char... chars2unescape)
    {
        String strUnescapedString = strEscapedString;
        for (char char2unescape : chars2unescape)
            strUnescapedString = strUnescapedString.replace("\\" + char2unescape, String.valueOf(char2unescape));

        return strUnescapedString;
    }







}
